/*
 * Copyright (c) 2001 VM Labs, Inc.  All rights reserved.
 *
 * NOTICE: VM Labs permits you to use, modify, and distribute this file
 * in accordance with the terms of the VM Labs license agreement
 * accompanying it. If you have received this file from a source other
 * than VM Labs, then your use, modification, or distribution of it
 * requires the prior written permission of VM Labs.
*/

//==================================================================================================
// Please widen your screen to fit 100 characters per line and set tab stops to 2 (if necessary)
//==================================================================================================
// 4567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
//==================================================================================================

//==================================================================================================
// Includes
//==================================================================================================

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <nuon/sprite2.h>
#include <nuon/termemu.h>
#include <nuon/bios.h>
#include <nuon/dma.h>
#include <nuon/video.h>
#include <nuon/joystick.h>
#include <nuon/nise.h>
#include <nuon/mutil.h>
#include <nuon/msprintf.h>

/*
    Level definitions
*/

#include "levels.h"
#include "custom.h"

// Bunch of constants

char versionString[] = "1.4E\0";
#define VERF_SEED 0  // Seed removed to discourage bogus high scores
#define DSI_DEBUG 0

#define TRACE_EXECUTION  0
#define DEBUG_PRINT      if (TRACE_EXECUTION) { printf("\nLine %d of %s", __LINE__, __FILE__); fflush(stdout); }
#define EMU_PRINT(x)     if (emulator) {  printf(x);  fflush(stdout);  }

#define SPEED_CHOMPER    0x1800  // Default chomper speed
#define SPEED_ENEMY      0x0800  // Starting enemy speed
#define SPEED_ENEMY_INC  0x0100  // How much to increment the enemy speed each level
#define SPEED_ENEMY_MAX  0x1800  // Maximum enemy speed
#define CHASE_EFFECT     0x8000  // How much eating a power pill affects enemies
#define SLOW_TILE_EFFECT 0x8000  // How much the 'slow' tiles affect the enemies
#define SPEED_PILL_BONUS 0x0400  // Consecutive pill bonus speeds you up

#define BONUS_PILL_AMT   10      // Number of pills to collect before bonus turns on

// Make pill bonus last long enough to cross two empty tiles with it on.  This should only allow
// one empty tile when the bonus is off.  The +3 is to account for truncation after division and 
// provide some cushion.
#define NO_PILL_DELAY    ((0x30000 / (SPEED_CHOMPER+SPEED_PILL_BONUS)) + 3)

#define SMOKE_LIFE       60      // Number of frames smoke lasts
#define GHOST_DEAD_TIME  180     // Number of frames ghosts stay dead while floating back
#define SCORE_TIME       60      // Number of frames the score sprite lasts (eating fruit, etc)
#define CHASE_TIME       500     // Number of frames power pill lasts

#define SCORE_PILL       10      // How many points things are...
#define SCORE_FRUIT      100
#define SCORE_POWERPILL  200
#define SCORE_GHOST12    400
#define SCORE_GHOST3456  800

#define TILE_EMPTY          ' '  // Characters used for maze construction...
#define TILE_START          'X'
#define TILE_WALL           '#'
#define TILE_PILL           '.'
#define TILE_POWERPILL      'o'
#define TILE_FRUIT          '*'
#define TILE_SLOW           '%'
#define TILE_SLOW_PILL      ':'
#define TILE_SLOW_POWERPILL '@'
#define TILE_SLOW_FRUIT     '$'
#define TILE_ENEMY_MIN      '1'
#define TILE_ENEMY_MAX      '6'
#define TILE_WALL_MIN       'A'
#define TILE_WALL_MAX       'P'

#define SCREEN_WIDTH  360
#define SCREEN_HEIGHT 240
#define DMAFLAGS     (DMA_PIXEL_WRITE | ((SCREEN_WIDTH/8)<<16) | (4<<4) | (1<<11))
#define DMAFLAGS_RD  (DMA_PIXEL_READ  | ((SCREEN_WIDTH/8)<<16) | (4<<4) | (1<<11))

#define DEADZONE_DIAGONAL 16     // Deadzone off the diagonals for analog control
#define DEADZONE_CENTER   32     // Deadzone off center for analog control

#define TILESIZE 8
#define INCG 1

#define TRUE  1
#define FALSE 0

#define OFFX 37
#define OFFY 20

#define SPRITESIZE  32
#define CHOMPERSIZE 40
#define ANIM_DELAY  2

#define GHOSTSCALE   0x6000
#define CHOMPERSCALE 0x6000
#define ITEMSCALE    0x6000
#define SCORESCALE   0x8000
#define LIVESSCALE   0x6000

#define MODE_START         0
#define MODE_SELECT        1
#define MODE_ARCADE        2
#define MODE_LEVEL_SELECT  3
#define MODE_TOURNAMENT    4
#define MODE_ARCADE_RAND   5
#define MODE_LEVEL_EDITOR  6

#define CHOMPER_ANIM_FRAMES 4

#define PILL_OFFSET_Y       0
#define POWERPILL_OFFSET_Y  1
#define BLANK_OFFSET_X      4
#define BLANK_OFFSET_Y      0
#define SLOW_OFFSET_X       4
#define SLOW_OFFSET_Y       1
#define SMOKE_OFFSET_X      0
#define SMOKE_OFFSET_Y      2

#define OFFSET_SHARP_X      0
#define OFFSET_SHARP_Y      6
#define OFFSET_STAR_X       1
#define OFFSET_STAR_Y       6
#define OFFSET_DOLLAR_X     2
#define OFFSET_DOLLAR_Y     6
#define OFFSET_COLON_X      3
#define OFFSET_COLON_Y      6
#define OFFSET_AT_X         4
#define OFFSET_AT_Y         6
#define OFFSET_O_X          5
#define OFFSET_O_Y          6

#define SCOREX 40
#define SCOREY 215

#define CHASEBARX 180
#define CHASEBARY 215

#define CHOMP_TILE_X  (chomper.tileX >> 16)
#define CHOMP_TILE_Y  (chomper.tileY >> 16)
#define CHOMP_MAZE_CONTENTS (maze[CHOMP_TILE_X + curLevel->w * CHOMP_TILE_Y])
#define mazeContents(x,y) (maze[(x) + (curLevel->w * (y))])
#define CHOMP_DIR_VERTICAL   ( chomper.angle & 0x4000)
#define CHOMP_DIR_HORIZONTAL (~chomper.angle & 0x4000)

#define SPRITE_SCREEN_X(x) (OFFX + curLevel->offsetX + ((((x) - 0x8000) * TILESIZE) >> 16))
#define SPRITE_SCREEN_Y(x) (OFFY + curLevel->offsetY + ((((x) - 0x8000) * TILESIZE) >> 16))

#define DELAY 15

#define NUM_SPRITES 16
#define NUM_BUFFERS 3
#define MAX_GHOSTS  6

//==================================================================================================
// Binaries (see data.s)
//==================================================================================================
extern char miscImages[];
extern char StartScreen[];
extern char GameOverScreen[];
extern char GameCompleteScreen[];
extern char LevelCompleteScreen[];

// Sounds
extern char GameBeginWav[];
extern char ChomperChompWav[];
extern char ChomperChomp2Wav[];
extern char FruitEatWav[];
extern char PowerEatWav[];
extern char ChaseWav[];
extern char GhostEatWav[];
extern char ChomperDiesWav[];
extern char ExtraChomperWav[];
extern char ClockTickWav[];
extern char GetReadyWav[];
extern char GoWav[];

extern char Font[];
extern char PauseImage[];
extern char ModeSelectImage[];
extern char ModeSelectTitleImage[];
extern char OptionsImage[];
extern char OptionsTitleImage[];
extern char LevelSelectImage[];
extern char NameEntryImage[];
extern char VerfCodeImage[];
extern char OnOffImage[];
extern char ClockImage[];
extern char LevelEditOptionImage[];
extern char LevelEditOtherImage[];
extern char LevelCharsImage[];

//==================================================================================================
// Structures for the ghosts and chomperman
//==================================================================================================

typedef struct
{
  SPRITE *spr;
  int gphase;
  int startX, startY;
  //  int ghostx, ghosty;
  //  int nghostx, nghosty;
  //  int ghostAx, ghostAy;
  int gincx, gincy;

  int tileX, tileY;  // 16.16 values indicating which tile Chomper is in (int=tile decimal=subtile)
  unsigned long speed;

  int way;
  int last_way;
  int hist[16], in, out;
  int active;
  int ghostNum;
  int immune;
} GHOST;


typedef struct
{
  SPRITE *chomperman;
  int aphase;
  int animDelay;
  int startX, startY;

  int tileX, tileY;  // 16.16 values indicating which tile Chomper is in (int=tile decimal=subtile)
  unsigned long speed;

  int angle;
  int direction, ndir;
  int consecutivePills;
  int noPillDelay;
  int bonus;
} CHOMPERMAN;

typedef struct {
  SPRITE *spr;
  int life;
} sSmoke;

//==================================================================================================
// Local prototypes
//==================================================================================================

int PlayWAV(void *wavaddr, int voice, PCMPOS *Pan, int volume, int echo);
void ResetGhost(GHOST *ghost);
void WaitForJoystick(int but);
int  PauseGame();	//SRA 1/14/2002 - pauses game, displays animated logo
void addSmoke();
void handleSmoke();
void resetSmoke();
inline int myPlayWAV(void *wavaddr, int voice, PCMPOS *Pan, int volume, int echo);
void checkAndLoadTGA(sCustomImg *img, int spritesX, int emuPrint);


//==================================================================================================
// Globals
//==================================================================================================

enum {
  kImgMisc,          kImgFont,         kImgLevelEditOption, kImgLevelEditOther,
  kImgLevelChars,    kImgStartScreen,  kImgGameOver,        kImgGameComplete,
  kImgLevelComplete, kImgPause,        kImgModeSelect,      kImgModeSelectTitle,
  kImgOptions,       kImgOptionsTitle, kImgLevelSelect,     kImgNameEntry,
  kImgVerfCode,      kImgOnOff,        kImgClock
};

// SDC 2002-02-10:  Broke up large sprite image into smaller pieces for customization
sCustomImg aImage[] = {
  { miscImages,           0 },
  { Font,                 0 },
  { LevelEditOptionImage, 0 },
  { LevelEditOtherImage,  0 },
  { LevelCharsImage,      0 },
  { StartScreen,          0 },
  { GameOverScreen,       0 },
  { GameCompleteScreen,   0 },
  { LevelCompleteScreen,  0 },
  { PauseImage,           0 },
  { ModeSelectImage,      0 },
  { ModeSelectTitleImage, 0 },
  { OptionsImage,         0 },
  { OptionsTitleImage,    0 },
  { LevelSelectImage,     0 },
  { NameEntryImage,       0 },
  { VerfCodeImage,        0 },
  { OnOffImage,           0 },
  { ClockImage,           0 },
};

PCMPOS Pan;

void *FrameBuffer[NUM_BUFFERS],*MazeBuf;
void *screenshotBuffer;
static int fps;
int drawBuffer = 0;
char *localScratch;
int localSize;
SPRITE *bonusSprite, *PauseSprite = NULL;
int saphase;

GHOST ghosts[MAX_GHOSTS];
CHOMPERMAN chomper;

int score;
int extra;
int lives;
int gameover;
int chase;
int pills;
int collected;
int delay;
int bPillBonus = 1;
int percentDelay;
int percent;

int numLevels = 0, numLevelsTournament = 0;
int curLevelIndex, levelTournament = 0;
LEVEL *curLevel;

int ghostsEaten;
int numGhosts;
int voiceChase, voiceClock;

int playerMax, wallMax, pillMax, fruitMax, ghostMax;
int playerIndex, wallIndex, pillIndex, fruitIndex, ghostIndex;
int debug;

// For high score verification
#define RAND_NAME_OFFSET    0
#define RAND_SCORE_OFFSET   (RAND_NAME_OFFSET+64)
#define RAND_LEVEL_OFFSET   (RAND_SCORE_OFFSET+32)
#define RAND_VERSION_OFFSET (RAND_LEVEL_OFFSET+32)

#define RAND_ARRAY_ENTRIES  (RAND_VERSION_OFFSET+32)
int aRand[256];
char curName[9] = "\0";

#define MAX_SMOKE 16
sSmoke aSmoke[MAX_SMOKE];
int smokeHead, smokeTail;

int *aRandMazes;

int gameMode;

int emulator = FALSE;
int mute = FALSE;
int initAudio = FALSE;

char maze[MAZE_MAX_W * MAZE_MAX_H];

//
// For DEBUG
//
int retVal;

/*
    Copy via DMA
*/


void CopyMaze(char *d,char *s,int pages)
{
  while(pages) {
    _DMALinear((64<<16)+(1<<13),s,localScratch);
    _DMALinear((64<<16),d,localScratch);
    pages--;
    s=&s[256];
    d=&d[256];
  }
}

/*
    Simple routine to draw a filled rectangle (via DMA)
*/

void myFillRect(char *screen, int dmaflags, int initx, int inity, int wide, int high, int color)
{
  int x, y;
  int lastx, lasty;
 
  lastx = initx + wide;
  lasty = inity + high;
  
  for (x = initx; x < lastx; x += 8)
    for (y = inity; y < lasty; y += 8)
      _raw_plotpixel(dmaflags, (void *) screen, (8<<16) | x, (8<<16) | y, color);
}

//==================================================================================================
//==================================================================================================

void screenshotCheck()
{
  int x, y;
  char *ptr;
  static int stillDown = 0;
  static int printed = 0;

  if ((_Controller[1].buttons & JOY_L) && (_Controller[1].buttons & JOY_R)) {
    if (DSI_DEBUG != 1) return;
    if (!printed) {
      screenshotBuffer = malloc(SCREEN_WIDTH * SCREEN_HEIGHT * 4);
      if (!screenshotBuffer) {
	printf("\nERROR: Could not allocate %d bytes for screenshot buffer",
	       SCREEN_WIDTH * SCREEN_HEIGHT * 4);
      }
      printf("\nBuffer address=%08X", (int) screenshotBuffer);  fflush(stdout);
      printed = 1;
    }

    if (!stillDown) {
      stillDown = 1;
      ptr = screenshotBuffer;
      for (y = 0; y < SCREEN_HEIGHT; y++)
	for (x = 0; x < SCREEN_WIDTH; x += 8) {
	  _DMABiLinear(DMAFLAGS_RD, FrameBuffer[drawBuffer], (8<<16)|(x<<0), (1<<16)|(y<<0),
		       localScratch);
	  _DMALinear((8<<16), ptr, localScratch);
	  ptr += 8*4;
	}
    }
  } else {
    stillDown = 0;
  }
}

//==================================================================================================
//==================================================================================================

void randomizeMazes()
{
  int i, j, levelIndex;

  aRandMazes = (int *) malloc (sizeof(int) * numLevels);
  for (i = 0; i < numLevels; i++) aRandMazes[i] = i;
  for (i = 0; i < numLevels; i++) {
    levelIndex = rand() % numLevels;
    j = aRandMazes[i];
    aRandMazes[i] = aRandMazes[levelIndex];
    aRandMazes[levelIndex] = j;
  }
}

//==================================================================================================
//==================================================================================================

void initMazes()
{
  int levelIndex;
  LEVEL *curLevel;
  int minScore;
  char *ptr;
  sLevelScore *levelScore;

  for (curLevel = levels, levelIndex = 0; strlen(curLevel->widthCalib); curLevel++, levelIndex++) {
    curLevel->w = strlen(curLevel->widthCalib);
    curLevel->h = strlen(curLevel->maze) / curLevel->w;
    curLevel->offsetX = (MAZE_MAX_W - curLevel->w) * (TILESIZE>>1);
    curLevel->offsetY = (MAZE_MAX_H - curLevel->h) * (TILESIZE>>1);
    numLevels++;
    curLevel->distance = 0;

    if (TRACE_EXECUTION) {
      printf("\nlevelIndex=%d  W=%d  H=%d", levelIndex, curLevel->w, curLevel->h);
      fflush(stdout);
    }
  }

  randomizeMazes();

  for (curLevel = aLevelTournament, levelIndex = 0; strlen(curLevel->widthCalib);
       curLevel++, levelIndex++) {
    curLevel->w = strlen(curLevel->widthCalib);
    curLevel->h = strlen(curLevel->maze) / curLevel->w;
    curLevel->offsetX = (MAZE_MAX_W - curLevel->w) * (TILESIZE>>1);
    curLevel->offsetY = (MAZE_MAX_H - curLevel->h) * (TILESIZE>>1);
    numLevelsTournament++;

    // Compute time it would take Chomper to traverse every non-wall tile
    curLevel->distance = 0;
    for (ptr = curLevel->maze; *ptr; ptr++)
      if ((*ptr != TILE_WALL) && ((*ptr < TILE_WALL_MIN) || (*ptr > TILE_WALL_MAX)))
	curLevel->distance += 0x10000;
    curLevel->distance /= SPEED_CHOMPER;

    if (TRACE_EXECUTION) {
      printf("\nlevelIndex=%d  W=%d  H=%d", levelIndex, curLevel->w, curLevel->h);
      fflush(stdout);
    }
  }

  aLevelScore = (sLevelScore *) malloc (numLevelsTournament * sizeof(sLevelScore));
  for (levelIndex = 0, levelScore = aLevelScore; levelIndex < numLevelsTournament;
       levelIndex++, levelScore++) {
    for (minScore = 0, ptr = aLevelTournament[levelIndex].maze; *ptr; ptr++) {
      switch (*ptr) {
        case TILE_PILL:
        case TILE_SLOW_PILL:       minScore += SCORE_PILL;       break;
        case TILE_POWERPILL:
        case TILE_SLOW_POWERPILL:  minScore += SCORE_POWERPILL;  break;
        case TILE_FRUIT:
        case TILE_SLOW_FRUIT:      minScore += SCORE_FRUIT;      break;
      }
    }
    levelScore->score   = minScore>>1;  // Assume a level-complete percentage of 50%
    levelScore->percent = 0;
    levelScore->name[0] = 0;
    levelScore->name[8] = 0;

    if (TRACE_EXECUTION) {
      printf("\nlevelIndex=%d  minScore=%d", levelIndex, minScore);
      fflush(stdout);
    }
  }
}

//
//
void updateTileChar(LEVEL *curLevel, int x, int y, long *frame)
{
  char mazeChar;
  int offsetX, offsetY;

  mazeChar = maze[x + (y * curLevel->w)];

  if ((mazeChar >= TILE_WALL_MIN) && (mazeChar <= TILE_WALL_MAX)) {
    offsetX = mazeChar - TILE_WALL_MIN;
  } else {
    switch(mazeChar) {
      case TILE_WALL:            offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 1;  break;
      case TILE_SLOW:            offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 2;  break;
      case TILE_SLOW_PILL:       offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 3;  break;
      case TILE_SLOW_POWERPILL:  offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 4;  break;
      case TILE_SLOW_FRUIT:      offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 5;  break;
      case TILE_PILL:            offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 6;  break;
      case TILE_POWERPILL:       offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 7;  break;
      case TILE_FRUIT:           offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 8;  break;

      case TILE_START:
	offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 9;
	chomper.startX = x;
	chomper.startY = y;
	break;

      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
	offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 10 + (mazeChar - TILE_ENEMY_MIN);
	ghosts[mazeChar - TILE_ENEMY_MIN].startX = x;
	ghosts[mazeChar - TILE_ENEMY_MIN].startY = y;
	if ((mazeChar - TILE_ENEMY_MIN + 1) > numGhosts)
	  numGhosts = mazeChar - TILE_ENEMY_MIN + 1;
	break;

      case TILE_EMPTY:
	offsetX = (TILE_WALL_MAX - TILE_WALL_MIN) + 10 + (TILE_ENEMY_MAX - TILE_ENEMY_MIN) + 1;
	break;

      default:
	offsetX = offsetY = 0;
	break;
    }
  }

  offsetY = offsetX >> 4;
  offsetX = offsetX & 0xF;
  checkAndLoadTGA(&aImage[kImgLevelChars], 1, TRUE);
  SPRBlitter(aImage[kImgLevelChars].spr.img, aImage[kImgLevelChars].spr.dmaflags,
             offsetX * 8, offsetY * 8, 8, 8, (char *) frame, DMAFLAGS,
	     curLevel->offsetX + (x * TILESIZE) + OFFX,
	     curLevel->offsetY + (y * TILESIZE) + OFFY,
	     0x10000, 0x10000, 0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
}

//
//
void updateTile(LEVEL *curLevel, int x, int y, long *frame,
		sCustomImg *imgWalls, sCustomImg *imgPills, sCustomImg *imgFruit,
		unsigned long wallScale, unsigned long pillScale, unsigned long fruitScale)
{
  char mazeChar, neighbor;
  int offsetX, offsetY;

  mazeChar = maze[x + (y * curLevel->w)];

  switch (mazeChar) {
    case TILE_SLOW:
    case TILE_SLOW_PILL:
    case TILE_SLOW_POWERPILL:
    case TILE_SLOW_FRUIT:
      SPRBlitter(imgWalls->spr.img, imgWalls->spr.dmaflags,
		 SLOW_OFFSET_X * imgWalls->w, SLOW_OFFSET_Y * imgWalls->w,
		 imgWalls->w, imgWalls->w,
		 (char *) frame, DMAFLAGS,
		 curLevel->offsetX + (x * TILESIZE) + OFFX,
		 curLevel->offsetY + (y * TILESIZE) + OFFY,
		 wallScale, wallScale,
		 0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
      break;

    default:
      // Clear the tile first with the blank tile
      SPRBlitter(imgWalls->spr.img, imgWalls->spr.dmaflags,
		 BLANK_OFFSET_X * imgWalls->w, BLANK_OFFSET_Y * imgWalls->w,
		 imgWalls->w, imgWalls->w,
		 (char *) frame, DMAFLAGS,
		 curLevel->offsetX + (x * TILESIZE) + OFFX,
		 curLevel->offsetY + (y * TILESIZE) + OFFY,
		 wallScale, wallScale,
		 0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
  }

  if ((mazeChar >= TILE_WALL_MIN) && (mazeChar <= TILE_WALL_MAX)) {
    SPRBlitter(imgWalls->spr.img, imgWalls->spr.dmaflags, 
	       (((mazeChar - TILE_WALL_MIN) >> 2) & 0x3) * imgWalls->w,
	       (((mazeChar - TILE_WALL_MIN) >> 0) & 0x3) * imgWalls->w,
	       SPRITESIZE / imgWalls->scale, SPRITESIZE / imgWalls->scale,
	       (char *) frame, DMAFLAGS,
	       curLevel->offsetX + (x * TILESIZE) + OFFX,
	       curLevel->offsetY + (y * TILESIZE) + OFFY,
	       wallScale, wallScale, 0, 0, 0,
	       SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
  }

  switch(mazeChar) {
    case TILE_WALL:
      offsetX = offsetY = 0;

      if (y != (curLevel->h - 1)) {
	neighbor = mazeContents(x,y+1);
	if ((neighbor == TILE_WALL) || ((neighbor>=TILE_WALL_MIN) && (neighbor<=TILE_WALL_MAX)))
	  offsetY+=1;
      }

      if (x != (curLevel->w-1)) {
	neighbor = mazeContents(x+1,y);
	if ((neighbor == TILE_WALL) || ((neighbor>=TILE_WALL_MIN) && (neighbor<=TILE_WALL_MAX)))
	  offsetY+=2;
      }

      if (y) {
	neighbor = mazeContents(x,y-1);
	if ((neighbor == TILE_WALL) || ((neighbor>=TILE_WALL_MIN) && (neighbor<=TILE_WALL_MAX)))
	  offsetX+=1;
      }

      if (x) {
	neighbor = mazeContents(x-1,y);
	if ((neighbor == TILE_WALL) || ((neighbor>=TILE_WALL_MIN) && (neighbor<=TILE_WALL_MAX)))
	  offsetX+=2;
      }
      SPRBlitter(imgWalls->spr.img, imgWalls->spr.dmaflags, 
		 offsetX * imgWalls->w,
		 offsetY * imgWalls->w,
		 SPRITESIZE / imgWalls->scale, SPRITESIZE / imgWalls->scale,
		 (char *) frame, DMAFLAGS,
		 curLevel->offsetX + (x * TILESIZE) + OFFX,
		 curLevel->offsetY + (y * TILESIZE) + OFFY,
		 wallScale, wallScale, 0, 0, 0,
		 SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
      break;

    case TILE_PILL:
    case TILE_SLOW_PILL:
      pills++;
      SPRBlitter(imgPills->spr.img, imgPills->spr.dmaflags,
		 0, PILL_OFFSET_Y * imgPills->w, imgPills->w, imgPills->w,
		 (char *) frame, DMAFLAGS,
		 curLevel->offsetX + (x * TILESIZE) + OFFX,
		 curLevel->offsetY + (y * TILESIZE) + OFFY,
		 pillScale, pillScale, 0, 0, 0,
		 SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
      break;
      
    case TILE_POWERPILL:
    case TILE_SLOW_POWERPILL:
      pills++;
      SPRBlitter(imgPills->spr.img, imgPills->spr.dmaflags,
		 0, POWERPILL_OFFSET_Y * imgPills->w,
		 imgPills->w, imgPills->w,
		 (char *) frame, DMAFLAGS,
		 curLevel->offsetX + (x * TILESIZE) + OFFX,
		 curLevel->offsetY + (y * TILESIZE) + OFFY,
		 pillScale, pillScale, 0, 0, 0,
		 SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
      break;

    case TILE_FRUIT:
    case TILE_SLOW_FRUIT:
      pills++;
      SPRBlitter(imgFruit->spr.img, imgFruit->spr.dmaflags, 0, 0,
		 imgFruit->w, imgFruit->w,
		 (char *) frame, DMAFLAGS,
		 curLevel->offsetX + (x * TILESIZE) + OFFX,
		 curLevel->offsetY + (y * TILESIZE) + OFFY,
		 fruitScale, fruitScale, 0, 0, 0,
		 SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
      break;

    case TILE_START:
      chomper.startX = x;
      chomper.startY = y;
      break;

    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
      ghosts[mazeChar - TILE_ENEMY_MIN].startX = x;
      ghosts[mazeChar - TILE_ENEMY_MIN].startY = y;
      if ((mazeChar - TILE_ENEMY_MIN + 1) > numGhosts) numGhosts = mazeChar - TILE_ENEMY_MIN + 1;
      break;
  }       
}

/*
    Convert ASCII description of a maze into a real framebuffer
*/

// SDC 2002-02-10:  Broke up sprite image file into separate images for further customization
//                  Rewrote '#' case to fit the new sprite layout
// SDC 2002-02-14:  Rewrote so that every tile gets drawn with "blank" tile first
// SDC 2002-05-26:  Removed restriction that mazes must be constant height/width
// SDC 2002-05-26:  Added slow tile

void CreateMaze(LEVEL *curLevel, sCustomImg *imgWalls, sCustomImg *imgPills, sCustomImg *imgFruit,
		long *frame)
{
  int x, y;
  unsigned long pillScale = 0, wallScale = 0, fruitScale = 0;

  pills = numGhosts = 0;
  chomper.startX = 18;
  chomper.startY = 19;
  if (imgWalls) wallScale  = (TILESIZE * 0x10000 * imgWalls->scale) / SPRITESIZE;
  if (imgPills) pillScale  = (TILESIZE * 0x10000 * imgPills->scale) / SPRITESIZE;
  if (imgFruit) fruitScale = (TILESIZE * 0x10000 * imgFruit->scale) / SPRITESIZE;

  for (x = 0; x < MAX_GHOSTS; x++) ghosts[x].startX = ghosts[x].startY = -1;

  for (x = 0; x < curLevel->w; x++)
    for (y = 0; y < curLevel->h; y++)
      if (!imgWalls || !imgPills || !imgFruit)
	updateTileChar(curLevel, x, y, frame);
      else
	updateTile(curLevel, x, y, frame, imgWalls, imgPills, imgFruit, 
		   wallScale, pillScale, fruitScale);
}

/*
  Erase a "pill" from maze and screen
*/

// SDC 2002-02-14:  Changed from erasing the tile to copying the blank tile over
// SDC 2002-05-26:  Added slow tile

void DeletePill(sCustomImg *img, long *frame, int x, int y)
{
  switch (CHOMP_MAZE_CONTENTS) {
    case TILE_SLOW:
    case TILE_SLOW_PILL:
    case TILE_SLOW_POWERPILL:
    case TILE_SLOW_FRUIT:
      SPRBlitter(img->spr.img, img->spr.dmaflags,
		 SLOW_OFFSET_X * img->w, SLOW_OFFSET_Y * img->w,
		 img->w, img->w,
		 (char *) frame, DMAFLAGS,
		 curLevel->offsetX + (x * TILESIZE) + OFFX,
		 curLevel->offsetY + (y * TILESIZE) + OFFY,
		 (TILESIZE * 0x10000 * img->scale) / SPRITESIZE,
		 (TILESIZE * 0x10000 * img->scale) / SPRITESIZE,
		 0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
      maze[x + (y * curLevel->w)] = '%';
      break;

    default:
      SPRBlitter(img->spr.img, img->spr.dmaflags,
		 BLANK_OFFSET_X * img->w, BLANK_OFFSET_Y * img->w,
		 img->w, img->w,
		 (char *) frame, DMAFLAGS,
		 curLevel->offsetX + (x * TILESIZE) + OFFX,
		 curLevel->offsetY + (y * TILESIZE) + OFFY,
		 (TILESIZE * 0x10000 * img->scale) / SPRITESIZE,
		 (TILESIZE * 0x10000 * img->scale) / SPRITESIZE,
		 0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
      maze[x + (y * curLevel->w)] = ' ';
      break;
  }
}

/*
 * wait for video
 * returns the number of elapsed fields
 * since the last sync
 */

int VideoSync(void)
{
  static int frames;
  static int lastfield;
  static int lastsecond;
  int ret;

  int curfield;

  // First check if we should capture a screenshot
  screenshotCheck();
    
  // we want to view the buffer we just finished drawing

  // set the video to point at the drawn buffer
  _VidChangeBase(VID_CHANNEL_MAIN, DMAFLAGS, FrameBuffer[drawBuffer]);
  
  // find out how many fields have been drawn
  curfield = _VidSync(-1);

  // remember that we've drawn a frame
  frames++;

  // have more than 60 fields (1 second) passed?
  if (curfield - lastsecond >= 60) {
    fps = frames;
    frames = 0;
    lastsecond = curfield;
  }

  // draw to a new buffer
  drawBuffer++;
  if (drawBuffer >= NUM_BUFFERS)
    drawBuffer = 0;

#if NUM_BUFFERS == 2
  // wait until a new field starts (so it's no longer showing the field we are going to draw on)
  curfield = _VidSync(0);
#endif

  ret = curfield - lastfield;
  while (ret == 0) {
    curfield = _VidSync(-1);
    ret = curfield - lastfield;
  }
  lastfield = curfield;
  return ret;
}

//==================================================================================================
// getFirstTGAPixel
//
// Parses a TGA file to find the first pixel color within the image.  Returns 0 (black) on failure,
// or the color upon success.
//
// Revision history
//   SDC 2002-02-14:  Initial revision
//==================================================================================================

int getFirstTGAPixel(char *tgaFile)
{
  int offsetColorMap;
  int offsetPixels;
  int colorMapStart;
  int colorMapLength;
  int colorBytes;
  unsigned char index;
  unsigned char r, g, b;

  offsetColorMap = offsetPixels = 18 + tgaFile[0];

  if (tgaFile[1] != 1) return(0);  // Must have color map
  if (tgaFile[2] != 9) return(0);  // Must be RLE color-mapped

  colorMapStart  = tgaFile[3] + (tgaFile[4] << 8);
  colorMapLength = tgaFile[5] + (tgaFile[6] << 8);

  if ((tgaFile[7] != 24) && (tgaFile[7] != 32)) return(0);  // Colors must be 24-bit or 32-bit
  if (tgaFile[16] != 8) return(0);  // Pixels must be 8-bit color depth

  colorBytes = (tgaFile[7] == 24) ? 3 : 4;
  offsetPixels += colorMapLength * colorBytes;

  index = tgaFile[offsetPixels+1];  // Skip past RLE opcode to get first color index
  r = tgaFile[offsetColorMap + (index*colorBytes) + 2];
  g = tgaFile[offsetColorMap + (index*colorBytes) + 1];
  b = tgaFile[offsetColorMap + (index*colorBytes) + 0];

  return((((int) r) << 24) | (((int) g) << 16) | (((int) b) << 8));
}

//==================================================================================================
// getImageScale
//
// The default sprite size is 32x32.  In the case of walls this gets scaled down to 8x8 for drawing
// the maze.  This function lets the program accept sprites of sizes other than 32x32 to save space.
// Returns a 16.16 scale value which can be used in the sprite functions.
//
// Revision history
//   SDC 2002-02-17:  Initial revision
//==================================================================================================

void setImageScale(sCustomImg *img, int spritesX)
{
  int width;

  width = img->tgaAddr[12] + (img->tgaAddr[13] << 8);
  width /= spritesX;

  switch (width) {
    case 8:
      img->scale = 4;
      img->w = 8;
      break;
    case 16:
      img->scale = 2;
      img->w = 16;
      break;
    default:
      img->scale = 1;
      img->w = 32;
      break;
  }
}

/*
    Start a score animation
*/

void StartScoreAnimation(sCustomImg *img, int s, int x, int y)
{
  int t;

  t = chomper.bonus ? 1 : 0;

  if (bonusSprite)
    SPRSetSpriteSource(bonusSprite,
		       (s * SPRITESIZE) / img->scale,
		       (t * SPRITESIZE) / img->scale,
		       img->w, img->w);
  else
    bonusSprite = SPRCreateSprite(&img->spr, (s * SPRITESIZE) / img->scale, 0,
				  img->w, img->w);
  
  SPRAddSprite(bonusSprite,
	       curLevel->offsetX + x * TILESIZE + OFFX,
	       curLevel->offsetY + y * TILESIZE + OFFY,
	       0,
	       SCORESCALE * img->scale, SCORESCALE * img->scale, 0x40000002, kBlack, 1);
  
  saphase = SCORE_TIME;             
}

/*
    Continue score animation
*/

void DoScoreAnimation(void)
{
  if (saphase) {
    SPRSetSpriteScale(bonusSprite, SCORESCALE + (SCORE_TIME - saphase) * 0x444,
		      SCORESCALE + (SCORE_TIME - saphase) * 0x444);
    SPRSetSpriteRotation(bonusSprite, (SCORE_TIME - saphase) * 0x800);
    saphase--;
    if (!saphase) SPRRemoveSprite(bonusSprite); 
  }
}

/*
    Start dead ghost animation
*/

void StartGhostDeadAnimation(GHOST *ghost)
{
  ghost->active = 0;
    
  //  ghost->ghostAx = (x * TILESIZE) + OFFX;
  //  ghost->ghostAy = (y * TILESIZE) + OFFY;
  ghost->gphase = GHOST_DEAD_TIME;
  ghost->gincx  = ((ghost->startX<<16) - ghost->tileX) / GHOST_DEAD_TIME;
  ghost->gincy  = ((ghost->startY<<16) - ghost->tileY) / GHOST_DEAD_TIME;
  SPRSetSpriteSource(ghost->spr, 6 * aGhostImg[ghostIndex].w, 0, 
		     aGhostImg[ghostIndex].w, aGhostImg[ghostIndex].w);
  SPRSetSpriteScale(ghost->spr, aGhostImg[ghostIndex].scale * 0xC000,
		    aGhostImg[ghostIndex].scale * 0xC000);
}

/*
    Control Chomperman (by joystick or remote control)
*/

// SDC 2002-02-10:  Changed control so no diagonals are allowed
//                  Made chomper sprite rotate so can reuse same animation for all 4 directions
// SDC 2002-02-15:  Added support for analog stick
// SDC 2002-03-03:  Began restructuring movement code
// SDC 2002-05-19:  Finished restructuring movement code
// SDC 2002-05-26:  Let Chomperman use warp tunnels

void ControlChomperman(void)
{
  int edge, joystick;
  static int old_joystick;
  char xAxis, yAxis;
  unsigned long prevTileX, prevTileY;
  int tileXInt, tileXFrac, tileYInt, tileYFrac;
  int tileXInt2, tileXFrac2, tileYInt2, tileYFrac2;
  int i;
        
  // logical OR between joystick and remote
  joystick = _Controller[1].buttons | _Controller[0].buttons;

  xAxis = _Controller[1].d1.xAxis;
  yAxis = _Controller[1].d2.yAxis;
  // If joystick is too close to diagonal directions then skip.  Prevents jitter from causing
  // direction to change unexpectedly.
  //
  // The diagonals occur when:  abs(xAxis) == abs(yAxis)
  //
  // To add some cushion, compute absolute difference between these two and ensure it is more
  // than a given deadzone.
  if (abs(abs(xAxis) - abs(yAxis)) > DEADZONE_DIAGONAL) {
    if (xAxis < -DEADZONE_CENTER) {
      if (yAxis < 0) joystick |= (-xAxis > -yAxis) ? CTRLR_DPAD_LEFT  : CTRLR_DPAD_DOWN;
      else           joystick |= (-xAxis >  yAxis) ? CTRLR_DPAD_LEFT  : CTRLR_DPAD_UP;

    } else if (xAxis > DEADZONE_CENTER) {
      if (yAxis < 0) joystick |= ( xAxis > -yAxis) ? CTRLR_DPAD_RIGHT : CTRLR_DPAD_DOWN;
      else           joystick |= ( xAxis >  yAxis) ? CTRLR_DPAD_RIGHT : CTRLR_DPAD_UP;

    } else if (yAxis < -DEADZONE_CENTER) {
      // Take advantage of the fact that previous "ifs" did not trigger (x-axis is within deadzone)
      joystick |= CTRLR_DPAD_DOWN;

    } else if (yAxis >  DEADZONE_CENTER) {
      // Take advantage of the fact that previous "ifs" did not trigger (x-axis is within deadzone)
      joystick |= CTRLR_DPAD_UP;
    }
  }

  // Clear out all but the direction bits
  joystick &= CTRLR_DPAD_UP | CTRLR_DPAD_DOWN | CTRLR_DPAD_LEFT | CTRLR_DPAD_RIGHT;

  edge = joystick ^ old_joystick;

  if (edge) chomper.ndir = joystick;

  prevTileX = chomper.tileX;
  prevTileY = chomper.tileY;
  switch (chomper.angle) {
    case 0x0000:
      if (chomper.direction) {
	chomper.tileX += chomper.speed;
	if (chomper.bonus) chomper.tileX += SPEED_PILL_BONUS;
      }

      // Chomper is facing right - no rotation or mirroring
      SPRSetSpriteRotation(chomper.chomperman, 0);
      SPRSetSpriteScale(chomper.chomperman,
			CHOMPERSCALE * aPlayerImg[playerIndex].scale,
			CHOMPERSCALE * aPlayerImg[playerIndex].scale);
      break;

    case 0x4000:
      if (chomper.direction) {
	chomper.tileY += chomper.speed;
	if (chomper.bonus) chomper.tileY += SPEED_PILL_BONUS;
      }

      // Chomper is facing up or down - rotate, but no mirroring
      SPRSetSpriteRotation(chomper.chomperman, chomper.angle);
      SPRSetSpriteScale(chomper.chomperman,
			CHOMPERSCALE * aPlayerImg[playerIndex].scale,
			CHOMPERSCALE * aPlayerImg[playerIndex].scale);
      break;

    case 0x8000:
      if (chomper.direction) {
	chomper.tileX -= chomper.speed;
	if (chomper.bonus) chomper.tileX -= SPEED_PILL_BONUS;
      }

      // Chomper is facing left - no rotation, but mirror x-direction
      SPRSetSpriteRotation(chomper.chomperman, 0);
      SPRSetSpriteScale(chomper.chomperman,
			-CHOMPERSCALE * aPlayerImg[playerIndex].scale,
			CHOMPERSCALE * aPlayerImg[playerIndex].scale);
      break;

    case 0xC000:
      if (chomper.direction) {
	chomper.tileY -= chomper.speed;
	if (chomper.bonus) chomper.tileY -= SPEED_PILL_BONUS;
      }

      // Chomper is facing up or down - rotate, but no mirroring
      SPRSetSpriteRotation(chomper.chomperman, chomper.angle);
      SPRSetSpriteScale(chomper.chomperman,
			CHOMPERSCALE * aPlayerImg[playerIndex].scale,
			CHOMPERSCALE * aPlayerImg[playerIndex].scale);
      break;
  }

  // Check if Chomper wrapped around the edge of the maze
  if (((chomper.tileX >> 16) != (prevTileX >> 16)) ||
      ((chomper.tileY >> 16) != (prevTileY >> 16))) {
    if (CHOMP_TILE_X < 0) {
      chomper.tileX += curLevel->w << 16;
    } else if (CHOMP_TILE_X >= curLevel->w) {
      chomper.tileX -= curLevel->w << 16;
    }

    if (CHOMP_TILE_Y < 0) {
      chomper.tileY += curLevel->h << 16;
    } else if (CHOMP_TILE_Y >= curLevel->h) {
      chomper.tileY -= curLevel->h << 16;
    }
  }

  tileXInt = chomper.tileX >> 16;
  tileYInt = chomper.tileY >> 16;
  tileXFrac = chomper.tileX & 0xFFFF;
  tileYFrac = chomper.tileY & 0xFFFF;

  tileXInt2 = prevTileX >> 16;
  tileYInt2 = prevTileY >> 16;
  tileXFrac2 = prevTileX & 0xFFFF;
  tileYFrac2 = prevTileY & 0xFFFF;

  // Check if Chomper ate the item at this location
  if ((tileXInt == tileXInt2) && (tileYInt == tileYInt2) &&
      (((tileXFrac <= 0x2000) && (tileXFrac2 > 0x2000)) ||
       ((tileXFrac >= 0x2000) && (tileXFrac2 < 0x2000)) ||
       ((tileYFrac <= 0x2000) && (tileYFrac2 > 0x2000)) ||
       ((tileYFrac >= 0x2000) && (tileYFrac2 < 0x2000)) ||

       ((tileXFrac <= 0xE000) && (tileXFrac2 > 0xE000)) ||
       ((tileXFrac >= 0xE000) && (tileXFrac2 < 0xE000)) ||
       ((tileYFrac <= 0xE000) && (tileYFrac2 > 0xE000)) ||
       ((tileYFrac >= 0xE000) && (tileYFrac2 < 0xE000)))) {

    if ((CHOMP_MAZE_CONTENTS == TILE_PILL) || (CHOMP_MAZE_CONTENTS == TILE_SLOW_PILL)) {
      //    printf("\nconsec=%d  delay=%d", chomper.consecutivePills, chomper.noPillDelay);
      if (chomper.bonus) addSmoke();
      if (chomper.bonus)
	myPlayWAV((void *) ChomperChomp2Wav, -1, &Pan, 0x40000000, 0x40000000);
      else
	myPlayWAV((void *) ChomperChompWav, -1, &Pan, 0x40000000, 0x40000000);
      DeletePill(&aWallImg[wallIndex], MazeBuf, CHOMP_TILE_X, CHOMP_TILE_Y);
      score += SCORE_PILL;
      if (chomper.bonus) score += SCORE_PILL >> 1;
      collected++;
      chomper.consecutivePills++;
      chomper.noPillDelay = NO_PILL_DELAY;
    }

    if ((CHOMP_MAZE_CONTENTS == TILE_FRUIT) || (CHOMP_MAZE_CONTENTS == TILE_SLOW_FRUIT)) {
      //    printf("\nconsec=%d  delay=%d", chomper.consecutivePills, chomper.noPillDelay);
      if (chomper.bonus) addSmoke();
      myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
      DeletePill(&aWallImg[wallIndex], MazeBuf, CHOMP_TILE_X, CHOMP_TILE_Y);
      score += SCORE_FRUIT;
      if (chomper.bonus) score += SCORE_FRUIT >> 1;
      StartScoreAnimation(&aImage[kImgMisc], 0, CHOMP_TILE_X, CHOMP_TILE_Y);
      collected++;
      chomper.consecutivePills++;
      chomper.noPillDelay = NO_PILL_DELAY;
    }

    if (((CHOMP_MAZE_CONTENTS == TILE_POWERPILL) ||
	 (CHOMP_MAZE_CONTENTS == TILE_SLOW_POWERPILL)) && !gameover) {
      //    printf("\nconsec=%d  delay=%d", chomper.consecutivePills, chomper.noPillDelay);
      if (chomper.bonus) addSmoke();
      myPlayWAV((void *) PowerEatWav, -1, &Pan, 0x40000000, 0x40000000);
      DeletePill(&aWallImg[wallIndex], MazeBuf, CHOMP_TILE_X, CHOMP_TILE_Y);
      score += SCORE_POWERPILL;
      if (chomper.bonus) score += SCORE_POWERPILL >> 1;
      StartScoreAnimation(&aImage[kImgMisc], 1, CHOMP_TILE_X, CHOMP_TILE_Y);
      if (!chase) voiceChase = myPlayWAV((void *) ChaseWav, 1, &Pan, 0x40000000, 0x40000000);

      // Reset this so the user cannot chain power pills together
      ghostsEaten = 0;

      chase = CHASE_TIME;
      collected++;
      chomper.consecutivePills++;
      chomper.noPillDelay = NO_PILL_DELAY;

      for (i = 0; i < MAX_GHOSTS; i++) ghosts[i].immune = 0;
    }

    if ((chomper.consecutivePills >= BONUS_PILL_AMT) &&
	(bPillBonus || (gameMode == MODE_TOURNAMENT))) {
      chomper.bonus = 1;
    }
  }

  // Turn off bonus if player went too long without eating a pill/fruit/powerpill
  chomper.noPillDelay--;
  //  printf("*");//%X,%X", chomper.tileX>>8, chomper.tileY>>8);
  if (chomper.noPillDelay <= 0)
    chomper.noPillDelay = chomper.consecutivePills = chomper.bonus = 0;

  // Check for change in direction when:
  // 1. Chomper is not moving
  // 2. Chomper just passed the center of the current tile
  // 3. Turning 180 degrees
  if (!chomper.direction ||

      ((tileXInt == tileXInt2) && (tileYInt == tileYInt2) &&
       (((tileXFrac <= 0x8000) && (tileXFrac2 > 0x8000)) ||
	((tileXFrac >= 0x8000) && (tileXFrac2 < 0x8000)) ||
	((tileYFrac <= 0x8000) && (tileYFrac2 > 0x8000)) ||
	((tileYFrac >= 0x8000) && (tileYFrac2 < 0x8000)))) ||

      (CHOMP_DIR_VERTICAL &&
       ((chomper.ndir == CTRLR_DPAD_UP  ) || (chomper.ndir == CTRLR_DPAD_DOWN ))) ||

      (CHOMP_DIR_HORIZONTAL &&
       ((chomper.ndir == CTRLR_DPAD_LEFT) || (chomper.ndir == CTRLR_DPAD_RIGHT))) ) {

    // Check if Chomperman changes direction
    if (chomper.ndir == CTRLR_DPAD_RIGHT) {
      if (mazeContents((CHOMP_TILE_X + 1), CHOMP_TILE_Y) != TILE_WALL) {
	if (CHOMP_DIR_VERTICAL) {
	  chomper.tileX += abs(chomper.tileY - ((chomper.tileY & 0xFFFF0000) | 0x8000));
	  chomper.tileY = (chomper.tileY & 0xFFFF0000) | 0x8000;
	}
	chomper.direction = CTRLR_DPAD_RIGHT;
      }
	  
    } else if (chomper.ndir == CTRLR_DPAD_LEFT) {
      if (mazeContents((CHOMP_TILE_X - 1), CHOMP_TILE_Y) != TILE_WALL) {
	if (CHOMP_DIR_VERTICAL) {
	  chomper.tileX -= abs(chomper.tileY - ((chomper.tileY & 0xFFFF0000) | 0x8000));
	  chomper.tileY = (chomper.tileY & 0xFFFF0000) | 0x8000;
	}
	chomper.direction = CTRLR_DPAD_LEFT;
      }
      
    } else if (chomper.ndir == CTRLR_DPAD_DOWN) {
      if (mazeContents(CHOMP_TILE_X, (CHOMP_TILE_Y + 1)) != TILE_WALL) {
	if (CHOMP_DIR_HORIZONTAL) {
	  chomper.tileY += abs(chomper.tileX - ((chomper.tileX & 0xFFFF0000) | 0x8000));
	  chomper.tileX = (chomper.tileX & 0xFFFF0000) | 0x8000;
	}
	chomper.direction = CTRLR_DPAD_DOWN;
      }
      
    } else if (chomper.ndir == CTRLR_DPAD_UP) {
      if (mazeContents(CHOMP_TILE_X, (CHOMP_TILE_Y - 1)) != TILE_WALL) {
	if (CHOMP_DIR_HORIZONTAL) {
	  chomper.tileY -= abs(chomper.tileX - ((chomper.tileX & 0xFFFF0000) | 0x8000));
	  chomper.tileX = (chomper.tileX & 0xFFFF0000) | 0x8000;
	}
	chomper.direction = CTRLR_DPAD_UP;
      }
    }
  }

  // Move Chomperman
  if (chomper.direction == CTRLR_DPAD_RIGHT) {
    if ((CHOMP_TILE_X < (curLevel->w-1)) &&
	(mazeContents((CHOMP_TILE_X + 1), CHOMP_TILE_Y) == TILE_WALL) &&
	((chomper.tileX & 0xFFFF) > 0x8000)) {
      chomper.direction = 0;
    } else {
      chomper.angle = 0;
    }

  } else if (chomper.direction == CTRLR_DPAD_LEFT) {
    if ((CHOMP_TILE_X > 0) &&
	(mazeContents((CHOMP_TILE_X - 1), CHOMP_TILE_Y) == TILE_WALL) &&
	((chomper.tileX & 0xFFFF) < 0x8000)) {
      chomper.direction = 0;
    } else {
      chomper.angle = 0x8000;
    }

  } else if (chomper.direction == CTRLR_DPAD_DOWN) {
    if ((CHOMP_TILE_Y < (curLevel->h-1)) &&
	(mazeContents(CHOMP_TILE_X, (CHOMP_TILE_Y + 1)) == TILE_WALL) &&
	((chomper.tileY & 0xFFFF) > 0x8000)) {
      chomper.direction = 0;
    } else {
      chomper.angle = 0x4000;
    }

  } else if (chomper.direction == CTRLR_DPAD_UP) {
    if ((CHOMP_TILE_Y > 0) &&
	(mazeContents(CHOMP_TILE_X, (CHOMP_TILE_Y - 1)) == TILE_WALL) &&
	((chomper.tileY & 0xFFFF) < 0x8000)) {
      chomper.direction = 0;
    } else {
      chomper.angle = 0xc000;
    }
  }

  // When Chomper is not moving center him in the tile
  if (!chomper.direction) {
    chomper.tileX = (chomper.tileX & 0xFFFF0000) | 0x8000;
    chomper.tileY = (chomper.tileY & 0xFFFF0000) | 0x8000;
  }

  // Put Chomperman in correct spot on screen
  // Subtract off half a tile since sprite is positioned by top-left pixel
  SPRSetSpriteXY(chomper.chomperman, SPRITE_SCREEN_X(chomper.tileX), SPRITE_SCREEN_Y(chomper.tileY));

  old_joystick = joystick;
}

/*
    Ghost AI
*/

// SDC 2002-05-26:  Make enemies affected by slow tile
// SDC 2002-05-26:  Let enemies use warp tunnels and be able to detect Chomp through them

void ControlGhost(GHOST *ghost)
{
  int x, y, c, visual, i;
  int ways, pw[4], sight[4];
  char savemaze[1024];            
  int destx, desty;
  unsigned long prevTileX, prevTileY;
  int prevWay;
  int ghostXNew, ghostYNew;
  char curMaze;

  prevTileX = ghost->tileX;
  prevTileY = ghost->tileY;
  prevWay   = ghost->way;

  ghost->tileX += ghost->gincx;
  ghost->tileY += ghost->gincy;

  if ((ghost->tileX >> 16) <  0          ) ghost->tileX += curLevel->w << 16;
  if ((ghost->tileX >> 16) >= curLevel->w) ghost->tileX -= curLevel->w << 16;
  if ((ghost->tileY >> 16) <  0          ) ghost->tileY += curLevel->h << 16;
  if ((ghost->tileY >> 16) >= curLevel->h) ghost->tileY -= curLevel->h << 16;

  if (ghost->gphase) {
    ghost->gphase--;
    SPRSetSpriteXY(ghost->spr, SPRITE_SCREEN_X(ghost->tileX), SPRITE_SCREEN_Y(ghost->tileY));
    if (!ghost->gphase) if (!ghost->active) ResetGhost(ghost);
    return;
  }
  
  // Ghost can turn only when it crosses a tile center
  if ((!ghost->gincx && !ghost->gincy) ||
      (((ghost->tileX & 0xFFFF0000) == (prevTileX & 0xFFFF0000)) &&
       ((ghost->tileY & 0xFFFF0000) == (prevTileY & 0xFFFF0000)) &&
       ((ghost->gincy && ((ghost->tileY ^ prevTileY) & 0x8000)) ||
        (ghost->gincx && ((ghost->tileX ^ prevTileX) & 0x8000))))) {

    ghost->gincx = ghost->gincy = 0;
      
    destx = CHOMP_TILE_X;
    desty = CHOMP_TILE_Y;
      
    if (chase && !ghost->immune) {
      destx = curLevel->w - destx;
      desty = curLevel->h - desty;
    }
      
    /* Do we have visual contact with chomperman?*/
    visual = 0;

    // Check above
    x = ghost->tileX >> 16;     
    y = ghost->tileY >> 16;     
    sight[3] = 0;
    do {
      sight[3]++;
      y--;
      if (y < 0) y += curLevel->h;
      c = maze[x + (y * curLevel->w)];
      if (c != TILE_WALL) {
	if ((y == desty) && (x == destx)) {
	  visual = 1;
	  ghost->way = 3;
	  break;
	}
      }
    } while ((c != TILE_WALL) && (y != (ghost->tileY >> 16)));

    // Check below
    if (!visual) {
      x = ghost->tileX >> 16;     
      y = ghost->tileY >> 16;     
      sight[2] = 0;
      do {
	sight[2]++;
	y++;
	if (y >= curLevel->h) y -= curLevel->h;
	c = maze[x + (y * curLevel->w)];
	if (c != TILE_WALL)	{
	  if ((y == desty) && (x == destx)) {
	    visual = 1;
	    ghost->way = 2;
	    break;
	  }
	}
      } while ((c != TILE_WALL) && (y != (ghost->tileY >> 16)));
    }

    // Check to left
    if (!visual) {
      x = ghost->tileX >> 16;     
      y = ghost->tileY >> 16;     
      sight[1] = 0;
      do {
	sight[1]++;
	x--;
	if (x < 0) x += curLevel->w;
	c = maze[x + (y * curLevel->w)];
	if (c != TILE_WALL) {
	  if ((y == desty) && (x == destx)) {
	    visual = 1;
	    ghost->way = 1;
	    break;
	  }
	}
      } while ((c != TILE_WALL) && (x != (ghost->tileX >> 16)));
    }

    // Check to right
    if (!visual) {
      x = ghost->tileX >> 16;     
      y = ghost->tileY >> 16;     
      sight[0] = 0;
      do {
	sight[0]++;
	x++;
	if (x >= curLevel->w) x -= curLevel->w;
	c = maze[x + (y * curLevel->w)];
	if (c != TILE_WALL) {
	  if ((y == desty) && (x == destx)) {
	    visual = 1;
	    ghost->way = 0;
	    break;
	  }
	}
      } while ((c != TILE_WALL) && (x != (ghost->tileX >> 16)));
    }
      
    //    if ((ghost->ghostNum == 0) && visual)
    //      printf("\n%d (  ,  ): VISUAL  way=%d", ghost->ghostNum, ghost->way);

    // Go into direction of ChomperMan if possible

    if (!visual) {
      i = (rand() % 3) + 1;
      x = ghost->tileX >> 16;     
      y = ghost->tileY >> 16;     
      if ((sight[0] > i) && (ghost->last_way != 1) && (destx > x)) {
	visual = 1;
	ghost->way = 0;
      }
      if ((sight[1] > i) && (ghost->last_way != 0) && (destx < x)) {
	visual = 1;
	ghost->way = 1;
      }
      if ((sight[2] > i) && (ghost->last_way != 3) && (desty > y)) {
	visual = 1;
	ghost->way = 2;
      }
      if ((sight[3] > i) && (ghost->last_way != 2) && (desty < y)) {
	visual = 1;
	ghost->way = 3;
      }

      //      if ((ghost->ghostNum == 0) && visual)
      //	printf("\n%d (%d,%d): TOWARDS way=%d", ghost->ghostNum, x, y, ghost->way);
    }

    // Go somewhere where we never have been (the last 16 moves)

    memcpy(savemaze, maze, curLevel->w * curLevel->h);
    if (!visual) {

      i = ghost->out;
      while(i != ghost->in) {
	maze[ghost->hist[i]] = TILE_WALL;
	i++;
	if (i == 16) i = 0;
      }
	  
      /* Now pick random direction*/
	  
      ways = 0;
      if (maze[(x+1) + ( y    * curLevel->w)] != TILE_WALL) pw[ways++] = 0;
      if (maze[(x-1) + ( y    * curLevel->w)] != TILE_WALL) pw[ways++] = 1;
      if (maze[ x    + ((y+1) * curLevel->w)] != TILE_WALL) pw[ways++] = 2;
      if (maze[ x    + ((y-1) * curLevel->w)] != TILE_WALL) pw[ways++] = 3;

      //      if (ghost->ghostNum == 0) for (i = 0; i < ways; i++) printf(" %d ", pw[i]);
	  
      memcpy(maze, savemaze, curLevel->w*curLevel->h);
      if (!ways) {
	if (maze[(x+1) + ( y    * curLevel->w)] != TILE_WALL) pw[ways++] = 0;
	if (maze[(x-1) + ( y    * curLevel->w)] != TILE_WALL) pw[ways++] = 1;
	if (maze[ x    + ((y+1) * curLevel->w)] != TILE_WALL) pw[ways++] = 2;
	if (maze[ x    + ((y-1) * curLevel->w)] != TILE_WALL) pw[ways++] = 3;
      }
	  
      ghost->way = pw[rand() % ways];
      //      if (ghost->ghostNum == 0)
      //	printf("\n%d (%d,%d): RANDOM  way=%d", ghost->ghostNum, x, y, ghost->way);
    }       
      
    ghostXNew = x = ghost->tileX >> 16;     
    ghostYNew = y = ghost->tileY >> 16;

    switch(ghost->way) {
	  
      case 0:
	curMaze = (x < (curLevel->w-1)) ? maze[(x+1)+y*curLevel->w] : maze[0+y*curLevel->w];
	if (curMaze != TILE_WALL) {
	  //	  if (ghost->ghostNum == 0) printf(" RIGHT ");
	  ghost->gphase = 1;
	  ghostXNew = (x < (curLevel->w-1)) ? (x+1) : 0;
	  ghost->gincx = ghost->speed;
	  if ((curMaze == TILE_SLOW          ) || (curMaze == TILE_SLOW_FRUIT) ||
	      (curMaze == TILE_SLOW_POWERPILL) || (curMaze == TILE_SLOW_PILL ))
	    ghost->gincx = (ghost->gincx * SLOW_TILE_EFFECT) >> 16;
	  if (chase && !ghost->immune) ghost->gincx = (ghost->gincx * CHASE_EFFECT) >> 16;
	} else
	  ghost->way     = -1;
	break;

      case 1:
	curMaze = (x > 0) ? maze[(x-1)+y*curLevel->w] : maze[(curLevel->w-1)+y*curLevel->w];
	if (curMaze != TILE_WALL) {
	  //	  if (ghost->ghostNum == 0) printf(" LEFT ");
	  ghost->gphase = 1;
	  ghostXNew = (x > 0) ? (x-1) : (curLevel->w-1);
	  ghost->gincx = -ghost->speed;
	  if ((curMaze == TILE_SLOW          ) || (curMaze == TILE_SLOW_FRUIT) ||
	      (curMaze == TILE_SLOW_POWERPILL) || (curMaze == TILE_SLOW_PILL ))
	    ghost->gincx = (ghost->gincx * SLOW_TILE_EFFECT) >> 16;
	  if (chase && !ghost->immune) ghost->gincx = (ghost->gincx * CHASE_EFFECT) >> 16;
	} else
	  ghost->way     = -1;
	break;
	
      case 2:
	curMaze = (y < (curLevel->h-1)) ? maze[x+(y+1)*curLevel->w] : maze[x+0*curLevel->w];
	if (curMaze != TILE_WALL) {
	  //	  if (ghost->ghostNum == 0) printf(" DOWN ");
	  ghost->gphase = 1;
	  ghostYNew = (y < (curLevel->h-1)) ? (y+1) : 0;
	  ghost->gincy = ghost->speed;
	  if ((curMaze == TILE_SLOW          ) || (curMaze == TILE_SLOW_FRUIT) ||
	      (curMaze == TILE_SLOW_POWERPILL) || (curMaze == TILE_SLOW_PILL ))
	    ghost->gincy = (ghost->gincy * SLOW_TILE_EFFECT) >> 16;
	  if (chase && !ghost->immune) ghost->gincy = (ghost->gincy * CHASE_EFFECT) >> 16;
	} else
	  ghost->way     = -1;
	break;
	
      case 3:
	curMaze = (y > 0) ? maze[x+(y-1)*curLevel->w] : maze[x+(curLevel->h-1)*curLevel->w];
	if (curMaze != TILE_WALL) {
	  //	  if (ghost->ghostNum == 0) printf(" UP ");
	  ghost->gphase = 1;
	  ghostYNew = (y > 0) ? (y-1) : (curLevel->h-1);
	  ghost->gincy = -ghost->speed;
	  if ((curMaze == TILE_SLOW          ) || (curMaze == TILE_SLOW_FRUIT) ||
	      (curMaze == TILE_SLOW_POWERPILL) || (curMaze == TILE_SLOW_PILL ))
	    ghost->gincy = (ghost->gincy * SLOW_TILE_EFFECT) >> 16;
	  if (chase && !ghost->immune) ghost->gincy = (ghost->gincy * CHASE_EFFECT) >> 16;
	} else
	  ghost->way     = -1;
	break;
    }

    //    if (ghost->ghostNum == 0) printf("  FINAL   way=%d", ghost->way);

    ghost->last_way = ghost->way;               
      
    i = ghost->in;
    //    if ((ghostXNew < 0) || (ghostYNew < 0)) {
    //      printf("\nghostXNew=%d ghostYNew=%d", ghostXNew, ghostYNew);
    //      fflush(stdout);
    //    }
    ghost->hist[i] = ghostXNew + (curLevel->w * ghostYNew);
    i++;
    if (i == 16) i = 0;
    if (i == ghost->out) {
      ghost->out++;
      if (ghost->out == 16) ghost->out = 0;
    }
    ghost->in = i;
  }

  // If ghost changed directions make sure it starts from the tile center
  if (ghost->way != prevWay) {
    ghost->tileX = (ghost->tileX & 0xFFFF0000) | 0x8000;
    ghost->tileY = (ghost->tileY & 0xFFFF0000) | 0x8000;
  }

  SPRSetSpriteXY(ghost->spr, SPRITE_SCREEN_X(ghost->tileX), SPRITE_SCREEN_Y(ghost->tileY));
}

void collisionDetect() {
  int i;
  int diffX, diffY;

  for(i = 0; i < numGhosts; i++)
    if (ghosts[i].active) {
      diffX = abs(ghosts[i].tileX - chomper.tileX);
      diffY = abs(ghosts[i].tileY - chomper.tileY);

      if ((diffX < 0x8000) && (diffY < 0x8000)) {
	if (!chase || ghosts[i].immune) {
	  myPlayWAV((void *) ChomperDiesWav, -1, &Pan, 0x40000000, 0x40000000);
	  gameover = 1;
	  if (chase) { if (!mute) PCMVoiceOff(voiceChase);  chase = 0; }
	} else if (chase && !ghosts[i].immune) {
	  ghosts[i].immune = 1;
	  ghostsEaten++;
	  myPlayWAV((void *) GhostEatWav, -1, &Pan, 0x40000000, 0x40000000);
	  if (ghostsEaten <= 2) {
	    score += SCORE_GHOST12;
	    if (chomper.bonus) score += SCORE_GHOST12 >> 1;
	  } else {
	    score += SCORE_GHOST3456;
	    if (chomper.bonus) score += SCORE_GHOST3456 >> 1;
	  }
	  StartScoreAnimation(&aImage[kImgMisc], (ghostsEaten > 2) ? 3 : 2, CHOMP_TILE_X, CHOMP_TILE_Y);
	  StartGhostDeadAnimation(&ghosts[i]);
	}
      }
    }
}

/*
  Initialize ghosts
*/


void ResetGhost(GHOST *ghost)
{
  ghost->active  = 1;
  ghost->way     = -1;

  ghost->gphase = ghost->gincx = ghost->gincy = 0;

  ghost->tileX = (ghost->startX << 16) | 0x8000;
  ghost->tileY = (ghost->startY << 16) | 0x8000;

  ghost->hist[0] = ghost->startX + (curLevel->w * ghost->startY);

  ghost->in = ghost->out = 0;
  if (SPRAddSprite(ghost->spr, SPRITE_SCREEN_X(ghost->tileX), SPRITE_SCREEN_Y(ghost->tileY), 0,
		   aGhostImg[ghostIndex].scale * GHOSTSCALE,
		   aGhostImg[ghostIndex].scale * GHOSTSCALE,
		   0x40000002, aGhostImg[ghostIndex].transColor, 1) < 0)
    SPRModifySprite(ghost->spr, SPRITE_SCREEN_X(ghost->tileX), SPRITE_SCREEN_Y(ghost->tileY), 0,
		    aGhostImg[ghostIndex].scale * GHOSTSCALE,
		    aGhostImg[ghostIndex].scale * GHOSTSCALE,
		    0x40000002, aGhostImg[ghostIndex].transColor, 1);
  SPRSetSpriteSource(ghost->spr, ghost->ghostNum * aGhostImg[ghostIndex].w, 0,
		     aGhostImg[ghostIndex].w, aGhostImg[ghostIndex].w);
}

/*
  Initialize all sprites
*/

void ResetSprites()
{
  int i;

  for(i = 0; i < numGhosts; i++) { ghosts[i].ghostNum = i;  ResetGhost(&ghosts[i]); }
  resetSmoke();
  
  chase = 0;
  chomper.direction = 0;
  chomper.tileX = (chomper.startX << 16) | 0x8000;  // Start at middle of tile
  chomper.tileY = (chomper.startY << 16) | 0x8000;
  chomper.speed = SPEED_CHOMPER;
  chomper.animDelay = ANIM_DELAY;
  chomper.aphase = 1;
  chomper.angle = 0;
  chomper.consecutivePills = 0;
  chomper.noPillDelay = 0;
  chomper.bonus = 0;

  if (SPRAddSprite(chomper.chomperman,
		   SPRITE_SCREEN_X(chomper.tileX),
		   SPRITE_SCREEN_Y(chomper.tileY),
		   0,
		   CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		   CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		   0x40000002, kBlack, 1) < 0)
    SPRModifySprite(chomper.chomperman,
		    SPRITE_SCREEN_X(chomper.tileX),
		    SPRITE_SCREEN_Y(chomper.tileY),
		    0,
		    CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		    CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		    0x40000002, kBlack, 1);

  SPRSetSpriteSource(chomper.chomperman, 
		     (chomper.aphase * SPRITESIZE) / aPlayerImg[playerIndex].scale, 0,
		     aPlayerImg[playerIndex].w, aPlayerImg[playerIndex].w);
}

/*
  Wait for a joystick (or remote) button
*/

void WaitForJoystick(int but)
{
  int edge,joystick;
  int old_joystick=0;
  
  do {
    joystick = _Controller[1].buttons | _Controller[0].buttons;
    edge = (joystick ^ old_joystick) & joystick; 
    old_joystick = joystick;
  } while (!(edge & but));
}

/*
  Draw Score
*/

void drawScore(int thisScore, char *frame, int x, int y, int trans, int printLives)
{
  char temp[256], *s;
  int offsetX, offsetY;
  
  //  sprintf(temp,"%d",fps);
  //  sprintf(temp,"%05d",chomper.noPillDelay);
  //  sprintf(temp,"%05d", percentDelay ? percentDelay : percent);
  sprintf(temp,"%05d", thisScore);

  s = temp;
  while(*s) {
    offsetX = (*s - '0') & 0x7;
    offsetY = 4 + ((*s - '0') >> 3);
    SPRBlitter(aImage[kImgFont].spr.img, aImage[kImgFont].spr.dmaflags,
               1 + offsetX*32, 1 + offsetY*32, 30, 30,
	       frame, DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0,
	       0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, kSpriteInterpolatedTrans | trans,
	       aImage[kImgFont].transColor);
    s++;
    x += (24 * SCORESCALE) / 0x10000;
  }
  
  if (printLives) {
    x += 24 * SCORESCALE / 0x10000;
    offsetX = lives & 0x7;
    offsetY = 4 + (lives >> 3);
    SPRBlitter(aImage[kImgFont].spr.img, aImage[kImgFont].spr.dmaflags,
               1 + offsetX*32, 1 + offsetY*32, 30, 30,
	       frame, DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0,
	       0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, aImage[kImgFont].transColor);

    x += 24 * SCORESCALE / 0x10000;
    offsetX = 2;
    offsetY = 5;
    SPRBlitter(aImage[kImgFont].spr.img, aImage[kImgFont].spr.dmaflags,
               offsetX*32, offsetY*32, 32, 32,
	       frame, DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0,
	       0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, aImage[kImgFont].transColor);

    x += 24 * SCORESCALE / 0x10000;
    SPRBlitter(aPlayerImg[playerIndex].spr.img, aPlayerImg[playerIndex].spr.dmaflags,
	       aPlayerImg[playerIndex].w, 0,
	       aPlayerImg[playerIndex].w, aPlayerImg[playerIndex].w,
	       frame, DMAFLAGS, x, y, 
	       LIVESSCALE * aPlayerImg[playerIndex].scale,
	       LIVESSCALE * aPlayerImg[playerIndex].scale,
	       0, 0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, kBlack);
  } else {
    if (percent) {
      if (percent >= curLevel->distance) {
	offsetX = 8;
      } else {
	offsetX = (percent<<3) / curLevel->distance;
      }

      x += 24 * SCORESCALE / 0x10000;
      checkAndLoadTGA(&aImage[kImgClock], 1, TRUE);
      SPRBlitter(aImage[kImgClock].spr.img, aImage[kImgClock].spr.dmaflags, offsetX*32, 0, 32, 32,
		 frame, DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0,
		 0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, 0);
    }
  }
}

//==================================================================================================
//==================================================================================================

int drawString(char *name, char *frame, int x, int y, int trans)
{
  char *ptr = name;
  int offsetX, offsetY;

  while(*ptr) {
    if ((toupper(*ptr) >= 'A') && (toupper(*ptr) <= 'Z')) {
      offsetX = (toupper(*ptr) - 'A') & 0x7;
      offsetY = (toupper(*ptr) - 'A') >> 3;
    } else if ((*ptr >= '0') && (*ptr <= '9')) {
      offsetX = (*ptr - '0') & 0x7;
      offsetY = 4 + ((*ptr - '0') >> 3);
    } else {
      switch (*ptr) {
        case ' ':  offsetX = 2;  offsetY = 3;  break;
        case '-':  offsetX = 3;  offsetY = 3;  break;
        case '\'': offsetX = 4;  offsetY = 3;  break;
        case '_':  offsetX = 5;  offsetY = 3;  break;
        case '!':  offsetX = 6;  offsetY = 3;  break;
        case '?':  offsetX = 7;  offsetY = 3;  break;
        case '.':  offsetX = 3;  offsetY = 5;  break;
        case '%':  offsetX = 7;  offsetY = 5;  break;
        default:   offsetX = 7;  offsetY = 3;  break;
      }
    }

    SPRBlitter(aImage[kImgFont].spr.img, aImage[kImgFont].spr.dmaflags,
               1 + (offsetX<<5), 1 + (offsetY<<5), 30, 30,
	       frame, DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0,
	       0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, kSpriteInterpolatedTrans | trans,
	       aImage[kImgFont].transColor);
    ptr++;
    x += (24 * SCORESCALE) / 0x10000;
  }

  return(x);
}

//==================================================================================================
//==================================================================================================

int PauseGame()
{
  checkAndLoadTGA(&aImage[kImgPause], 1, TRUE);
  if (!PauseSprite) PauseSprite = SPRCreateSprite(&aImage[kImgPause].spr, 0, 0, 256, 144);
  SPRAddSprite (PauseSprite, SCREEN_WIDTH/2, SCREEN_HEIGHT/2, 0, 0x9000, 0x7000, 0x40000002, 0, 1);
  SPRDraw (0,1);
  VideoSync();
  //	_VidChangeBase(VID_CHANNEL_MAIN,DMAFLAGS,FrameBuffer[0]);
  if (!mute) PCMVoiceOff(voiceChase);
  WaitForJoystick(JOY_A);
  if (chase) voiceChase = myPlayWAV((void *) ChaseWav, 1, &Pan, 0x40000000, 0x40000000);
  SPRRemoveSprite(PauseSprite);
  if ((_Controller[1].buttons | _Controller[0].buttons) & JOY_B) return(1);
  return(0);
}

//==================================================================================================
//==================================================================================================

void addSmoke()
{
  int newHead;

  newHead = smokeHead + 1;
  if (newHead >= MAX_SMOKE) newHead = 0;
  if (newHead == smokeTail) return;

  SPRAddSprite(aSmoke[smokeHead].spr, SPRITE_SCREEN_X(chomper.tileX),SPRITE_SCREEN_Y(chomper.tileY),
	       rand() & 0xFFFF, 0x4000, 0x4000, kSpriteInterpolatedTrans | (1<<29), 0, 2);
  aSmoke[smokeHead].life = 0;

  smokeHead++;
  if (smokeHead >= MAX_SMOKE) smokeHead = 0;
}

//==================================================================================================
//==================================================================================================

void handleSmoke()
{
  int i;

  for (i = smokeTail; i != smokeHead; i = (i >= (MAX_SMOKE-1)) ? 0 : i + 1) {

    aSmoke[i].life++;
    aSmoke[i].spr->xscale += 0x0100;
    aSmoke[i].spr->yscale += 0x0100;
    aSmoke[i].spr->trans -= 1<<22;

    if (aSmoke[i].life >= SMOKE_LIFE) {
      if (i != smokeTail) {
	printf("\ni=%d  smokeTail=%d  smokeHead=%d", i, smokeTail, smokeHead);
	fflush(stdout);
      }
      SPRRemoveSprite(aSmoke[i].spr);
      smokeTail++;
      if (smokeTail >= MAX_SMOKE) smokeTail = 0;
    }

  }
}

//==================================================================================================
//==================================================================================================

void resetSmoke()
{
  int i;

  for (i = smokeTail; i != smokeHead; i = (i >= (MAX_SMOKE-1)) ? 0 : i + 1)
    SPRRemoveSprite(aSmoke[i].spr);

  smokeHead = smokeTail = 0;
}

//==================================================================================================
//==================================================================================================

void selectPlayer()
{
  int done = 0;
  int curSelection = playerIndex;
  int x, y, prevX, prevY;
  int pinc = 1;
  int chomperWidth;
  int joystick = 0, prevJoystick;
  int prevPlayer = playerIndex;

  prevX = chomper.chomperman->x;
  prevY = chomper.chomperman->y;
  x = SCREEN_WIDTH/2;
  y = SCREEN_HEIGHT - 40;

  chomperWidth = (CHOMPERSCALE * SPRITESIZE) / 0x10000;
  chomperWidth = (chomperWidth >> 1) + 1;
  chomper.aphase = 1;

  while (!done) {
    myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, prevX - chomperWidth, prevY - chomperWidth,
	       chomperWidth*2, chomperWidth*2, kBlack);
    myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, x - chomperWidth, y - chomperWidth,
	       chomperWidth*2, chomperWidth*2, kBlack);

    // logical OR between joystick and remote
    prevJoystick = joystick;
    joystick = _Controller[1].buttons | _Controller[0].buttons;

    if (_Controller[1].d2.yAxis >  DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_UP;
    if (_Controller[1].d2.yAxis < -DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_DOWN;

    if ((joystick & CTRLR_DPAD_UP) && !(prevJoystick & CTRLR_DPAD_UP)) {
      myPlayWAV((void *) ChomperChomp2Wav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection--;
    }

    if ((joystick & CTRLR_DPAD_DOWN) && !(prevJoystick & CTRLR_DPAD_DOWN)) {
      myPlayWAV((void *) ChomperChomp2Wav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection++;
    }

    if (curSelection <          0) curSelection += playerMax;
    if (curSelection >= playerMax) curSelection -= playerMax;

    if (curSelection != playerIndex) {
      playerIndex = curSelection;
      SPRRemoveSprite(chomper.chomperman);
      SPRDeleteSprite(chomper.chomperman);
      chomper.chomperman = SPRCreateSprite(&aPlayerImg[playerIndex].spr,
					   aPlayerImg[playerIndex].w, 0,
					   aPlayerImg[playerIndex].w,
					   aPlayerImg[playerIndex].w);
      SPRAddSprite(chomper.chomperman, x, y, 0,
		   CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		   CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		   0x40000002, kBlack, 1);
    }

    SPRSetSpriteXY(chomper.chomperman, x, y);

    if (!chomper.animDelay) {
      SPRSetSpriteSource(chomper.chomperman, 
			 (chomper.aphase * SPRITESIZE) / aPlayerImg[playerIndex].scale, 0,
			 aPlayerImg[playerIndex].w, aPlayerImg[playerIndex].w);
      chomper.aphase += pinc;
      if ((chomper.aphase == 0) || (chomper.aphase == (CHOMPER_ANIM_FRAMES-1))) pinc = -pinc;
      chomper.animDelay = ANIM_DELAY;
    } else {
      chomper.animDelay--;
    }

    if (joystick & JOY_A) done = 1;
    if (joystick & JOY_B) { done = 1;  curSelection = prevPlayer; }

    // Process display list
    SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH - 1,
		     SCREEN_HEIGHT - 1, kBlack);
    SPRDraw(0,0);
    SPRWait();
    VideoSync();
  }

  if (curSelection != playerIndex) {
    playerIndex = curSelection;
    SPRRemoveSprite(chomper.chomperman);
    SPRDeleteSprite(chomper.chomperman);
    chomper.chomperman = SPRCreateSprite(&aPlayerImg[playerIndex].spr,
					 aPlayerImg[playerIndex].w,
					 0,
					 aPlayerImg[playerIndex].w,
					 aPlayerImg[playerIndex].w);
    SPRAddSprite(chomper.chomperman, x, y, 0,
		 CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		 CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		 0x40000002, kBlack, 1);
  }

  while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));
  chomper.aphase = 1;
}
      
//==================================================================================================
//==================================================================================================

void getOptions()
{
  int done = 0;
  int curSelection = 0;
  int x, y, menuX, menuY;
  int pinc = 1;
  int joystick = 0, prevJoystick;

  chomper.aphase = 1;
  
  while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));

  while (!done) {
    myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);

    x = SCREEN_WIDTH/2;
    y = SCREEN_HEIGHT/2 - 50;
    checkAndLoadTGA(&aImage[kImgOptionsTitle], 1, TRUE);
    SPRBlitter(aImage[kImgOptionsTitle].spr.img, aImage[kImgOptionsTitle].spr.dmaflags,
               0, 0, 160, 64, FrameBuffer[drawBuffer], DMAFLAGS, x, y,
	       0x10000 * SCREEN_WIDTH / 720, 0x10000 * SCREEN_HEIGHT / 480,
	       0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);

    menuX = SCREEN_WIDTH/2 + 15;
    menuY = SCREEN_HEIGHT/2 + 22;
    checkAndLoadTGA(&aImage[kImgOptions], 1, TRUE);
    SPRBlitter(aImage[kImgOptions].spr.img, aImage[kImgOptions].spr.dmaflags, 0, 0, 192, 192,
	       FrameBuffer[drawBuffer], DMAFLAGS, menuX, menuY,
	       0x10000 * SCREEN_WIDTH / 720, 0x10000 * SCREEN_HEIGHT / 480,
	       0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);

    x = menuX + 50;
    y = menuY - 36;
    checkAndLoadTGA(&aImage[kImgOnOff], 1, TRUE);
    SPRBlitter(aImage[kImgOnOff].spr.img, aImage[kImgOnOff].spr.dmaflags, 64 * bPillBonus,
               0, 64, 64, FrameBuffer[drawBuffer], DMAFLAGS, x, y,
	       0x10000 * SCREEN_WIDTH / 720, 0x10000 * SCREEN_HEIGHT / 480,
	       0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
    x = menuX + 20;
    y = menuY + 7;
    SPRBlitter(aImage[kImgOnOff].spr.img, aImage[kImgOnOff].spr.dmaflags, 64 * mute,
               0, 64, 64, FrameBuffer[drawBuffer], DMAFLAGS, x, y,
	       0x10000 * SCREEN_WIDTH / 720, 0x10000 * SCREEN_HEIGHT / 480,
	       0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);

    // logical OR between joystick and remote
    prevJoystick = joystick;
    joystick = _Controller[1].buttons | _Controller[0].buttons;

    if (_Controller[1].d2.yAxis >  DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_UP;
    if (_Controller[1].d2.yAxis < -DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_DOWN;

    if ((curSelection > 0) && (joystick & CTRLR_DPAD_UP) && !(prevJoystick & CTRLR_DPAD_UP)) {
      myPlayWAV((void *) ChomperChompWav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection--;
    }

    if ((curSelection < 3) && (joystick & CTRLR_DPAD_DOWN) && !(prevJoystick & CTRLR_DPAD_DOWN)) {
      myPlayWAV((void *) ChomperChompWav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection++;
    }

    x = SCREEN_WIDTH/2  - 45;
    y = SCREEN_HEIGHT/2 + 10;
    switch (curSelection) {
      case 0:  y -= 24;  break;
      case 1:  y -= 2;   break;
      case 2:  y += 20;  break;
      case 3:  y += 51;  break;
    }

    SPRSetSpriteXY(chomper.chomperman, x, y);

    if (!chomper.animDelay) {
      SPRSetSpriteSource(chomper.chomperman, 
			 (chomper.aphase * SPRITESIZE) / aPlayerImg[playerIndex].scale, 0,
			 aPlayerImg[playerIndex].w, aPlayerImg[playerIndex].w);
      chomper.aphase += pinc;
      if ((chomper.aphase == 0) || (chomper.aphase == (CHOMPER_ANIM_FRAMES-1))) pinc = -pinc;
      chomper.animDelay = ANIM_DELAY;
    } else {
      chomper.animDelay--;
    }

    // Process display list
    SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH - 1,
		     SCREEN_HEIGHT - 1, kBlack);
    SPRDraw(0,0);
    SPRWait();
    VideoSync();

    if (joystick & JOY_A) {
      while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));

      switch (curSelection) {
        case 0:
	  myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
	  bPillBonus = 1 - bPillBonus;
	  break;
        case 1:
          selectPlayer();
          break;
        case 2:
	  myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
          mute = 1 - mute;
          if (!mute && !initAudio) { initAudio = TRUE;  retVal = AUDIOInit(); }
	  myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
          break;
        case 3:
          done = 1;
          break;
      }
    } else if (joystick & JOY_B) {
      while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));

      done = 1;
    }
  }

  myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);

  while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));
}
      
//==================================================================================================
//==================================================================================================

int getGameMode()
{
  static int prevSelection = 0;
  int done = 0;
  int curSelection = prevSelection;
  int x, y;
  int pinc = 1;
  int joystick = 0, prevJoystick;

  SPRAddSprite(chomper.chomperman,
	       CHOMP_TILE_X * TILESIZE + OFFX, CHOMP_TILE_Y * TILESIZE + OFFY,
	       0,
	       CHOMPERSCALE * aPlayerImg[playerIndex].scale,
	       CHOMPERSCALE * aPlayerImg[playerIndex].scale,
	       0x40000002, kBlack, 1);
  chomper.aphase = 1;
  
  while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));

  while (!done) {
    myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);

    x = SCREEN_WIDTH/2;
    y = SCREEN_HEIGHT/2 - 85;
    checkAndLoadTGA(&aImage[kImgModeSelectTitle], 1, TRUE);
    SPRBlitter(aImage[kImgModeSelectTitle].spr.img, aImage[kImgModeSelectTitle].spr.dmaflags,
               0, 0, 360, 64, FrameBuffer[drawBuffer], DMAFLAGS, x, y,
	       0x10000 * SCREEN_WIDTH / 720, 0x10000 * SCREEN_HEIGHT / 480,
	       0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);

    x = SCREEN_WIDTH/2 + 15;
    y = SCREEN_HEIGHT/2 + 16;
    checkAndLoadTGA(&aImage[kImgModeSelect], 1, TRUE);
    SPRBlitter(aImage[kImgModeSelect].spr.img, aImage[kImgModeSelect].spr.dmaflags, 0, 0, 256, 320,
	       FrameBuffer[drawBuffer], DMAFLAGS, x, y,
	       0x10000 * SCREEN_WIDTH / 720, 0x10000 * SCREEN_HEIGHT / 480,
	       0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);

    // logical OR between joystick and remote
    prevJoystick = joystick;
    joystick = _Controller[1].buttons | _Controller[0].buttons;

    if (_Controller[1].d2.yAxis >  DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_UP;
    if (_Controller[1].d2.yAxis < -DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_DOWN;

    if ((curSelection > 0) && (joystick & CTRLR_DPAD_UP) && !(prevJoystick & CTRLR_DPAD_UP)) {
      myPlayWAV((void *) ChomperChompWav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection--;
    }

    if ((curSelection < 5) && (joystick & CTRLR_DPAD_DOWN) && !(prevJoystick & CTRLR_DPAD_DOWN)) {
      myPlayWAV((void *) ChomperChompWav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection++;
    }

    x -= 75;
    y = 67;
    switch (curSelection) {
      case 0:  y += 0; break;
      case 1:  y += 20; break;
      case 2:  y += 40;  break;
      case 3:  y += 72; break;
      case 4:  y += 100; break;
      case 5:  y += 129; break;
    }

    SPRSetSpriteXY(chomper.chomperman, x, y);

    if (!chomper.animDelay) {
      SPRSetSpriteSource(chomper.chomperman, 
			 (chomper.aphase * SPRITESIZE) / aPlayerImg[playerIndex].scale, 0,
			 aPlayerImg[playerIndex].w, aPlayerImg[playerIndex].w);
      chomper.aphase += pinc;
      if ((chomper.aphase == 0) || (chomper.aphase == (CHOMPER_ANIM_FRAMES-1))) pinc = -pinc;
      chomper.animDelay = ANIM_DELAY;
    } else {
      chomper.animDelay--;
    }

    // Process display list
    SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH - 1,
		     SCREEN_HEIGHT - 1, kBlack);
    SPRDraw(0,0);
    SPRWait();
    VideoSync();

    if (joystick & JOY_A) {
      switch (curSelection) {
        case 0:
        case 1:
        case 2:
        case 3:
        case 4:  done = 1;  break;

        case 5:
	  myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
	  getOptions();
	  chomper.aphase = 1;
	  break;
      }
    }
  }

  prevSelection = curSelection;

  while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));

  SPRRemoveSprite(chomper.chomperman);

  switch (curSelection) {
    case 0:  return(MODE_ARCADE);
    case 1:  return(MODE_ARCADE_RAND);
    case 2:  return(MODE_LEVEL_SELECT);
    case 3:  return(MODE_TOURNAMENT);
    case 4:  return(MODE_LEVEL_EDITOR);
  }
  return(MODE_SELECT);
}
      
//==================================================================================================
//==================================================================================================

int getVerfCode(int score, char *name, int level)
{
  char *ptr;
  int charIndex, bit;
  int code;
  int debug = 0;

  code = score;

  if (debug) printf("\nINIT Code=%08X", code);

  for (ptr = name, charIndex = 0; *ptr; ptr++, charIndex++)
    for (bit = 0; bit < 8; bit++) {
      if (*ptr & (1<<bit)) {
	if (debug) printf("\nNAME %c %d:  before=%08X  ", *ptr, bit, code);
	code ^= aRand[RAND_NAME_OFFSET + charIndex*8 + bit];
	if (debug)
	  printf("field=%08X  after=%08X  ", aRand[RAND_NAME_OFFSET + charIndex*8 + bit], code);
      }
    }

  for (bit = 0; bit < 32; bit++)
    if (score & (1<<bit)) {
      if (debug) printf("\nSCORE %d:  before=%08X  ", bit, code);
      code ^= aRand[RAND_SCORE_OFFSET + bit];
      if (debug) printf("field=%08X  after=%08X  ", aRand[RAND_SCORE_OFFSET + bit], code);
    }

  for (bit = 0; bit < 32; bit++)
    if (level & (1<<bit)) {
      if (debug) printf("\nLEVEL %d:  before=%08X  ", bit, code);
      code ^= aRand[RAND_LEVEL_OFFSET + bit];
      if (debug) printf("field=%08X  after=%08X  ", aRand[RAND_LEVEL_OFFSET + bit], code);
    }

  for (ptr = versionString, charIndex = 0; *ptr; ptr++, charIndex++)
    for (bit = 0; bit < 8; bit++) {
      if (*ptr & (1<<bit)) {
	if (debug) printf("\nVERSION %c %d:  before=%08X  ", *ptr, bit, code);
	code ^= aRand[RAND_VERSION_OFFSET + charIndex*8 + bit];
	if (debug)
	  printf("field=%08X  after=%08X  ", aRand[RAND_VERSION_OFFSET + charIndex*8 + bit], code);
      }
    }

  fflush(stdout);
  return(code);
}

//==================================================================================================
//==================================================================================================

int selectLevel(LEVEL *aCurLevels, int levelIndex, int maxLevels)
{
  int done = 0;
  int curSelection = levelIndex;
  int joystick = 0, prevJoystick;
  int newMaze = 1, randomize = 1;
  int i;
  int x, y;
  int levelNum;
  int blink = 1;
  int highScoreTimeout = 0;
  int offsetX, offsetY;
  int verfCode;
  char tmpString[12];

  if (levelIndex < 0) curSelection = levelIndex = 0;

  SPRAddSprite(chomper.chomperman, 0, 0, 0,
	       CHOMPERSCALE * aPlayerImg[playerIndex].scale,
	       CHOMPERSCALE * aPlayerImg[playerIndex].scale,
	       0x40000002, kBlack, 1);

  while (!done) {

    // Draw a new maze?
    if (newMaze) {

      // Get random graphics?
      if (randomize) {
        if (emulator) {
          wallIndex  = pillIndex  = fruitIndex = ghostIndex = 0;
        } else {
          wallIndex  = rand() % wallMax;
          pillIndex  = rand() % pillMax;
          fruitIndex = rand() % fruitMax;
          ghostIndex = rand() % ghostMax;
        }
	randomize = 0;
      }

      highScoreTimeout = 120;

      // Make sure all sprites are properly installed and initialized
      EMU_PRINT("\nLoading level images: wall");
      checkAndLoadTGA(&aWallImg [wallIndex ], 5, FALSE);
      EMU_PRINT(" pill");
      checkAndLoadTGA(&aPillImg [pillIndex ], 1, FALSE);
      EMU_PRINT(" fruit");
      checkAndLoadTGA(&aFruitImg[fruitIndex], 1, FALSE);
      EMU_PRINT(" ghost");
      checkAndLoadTGA(&aGhostImg[ghostIndex], 7, FALSE);

      // Draw the maze
      EMU_PRINT("\nCreating maze...");
      curLevel = &aCurLevels[curSelection];
      memcpy(maze, curLevel->maze, curLevel->w * curLevel->h);
      myFillRect(MazeBuf, DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);
      CreateMaze(curLevel, &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex],
		 MazeBuf);
      newMaze = 0;
      EMU_PRINT("done");

      // Need to delete and recreate ghosts since sprite changes every level
      for(i = 0; i < MAX_GHOSTS; i++) {
	if (ghosts[i].spr) {
	  SPRRemoveSprite(ghosts[i].spr);
	  SPRDeleteSprite(ghosts[i].spr);
	}
	ghosts[i].spr = SPRCreateSprite(&aGhostImg[ghostIndex].spr,
					i * aGhostImg[ghostIndex].w, 0,
					aGhostImg[ghostIndex].w, aGhostImg[ghostIndex].w);
      }

      for(i = 0; i < MAX_GHOSTS; i++) {
	if (ghosts[i].startX >= 0) {
	  ghosts[i].tileX  = (ghosts[i].startX << 16) | 0x8000;
	  ghosts[i].tileY  = (ghosts[i].startY << 16) | 0x8000;
	  SPRAddSprite(ghosts[i].spr,
		       SPRITE_SCREEN_X(ghosts[i].tileX),
		       SPRITE_SCREEN_Y(ghosts[i].tileY),
		       0,
		       aGhostImg[ghostIndex].scale * GHOSTSCALE,
		       aGhostImg[ghostIndex].scale * GHOSTSCALE,
		       0x40000002, aGhostImg[ghostIndex].transColor, 1);
	}
      }

      chomper.tileX = (chomper.startX << 16) | 0x8000;
      chomper.tileY = (chomper.startY << 16) | 0x8000;
      chomper.aphase = 1;
      SPRSetSpriteXY(chomper.chomperman,
		     SPRITE_SCREEN_X(chomper.tileX),
		     SPRITE_SCREEN_Y(chomper.tileY));
      SPRSetSpriteSource(chomper.chomperman,
			 (chomper.aphase * SPRITESIZE) / aPlayerImg[playerIndex].scale, 0,
			 aPlayerImg[playerIndex].w, aPlayerImg[playerIndex].w);
    }

    myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);

    CopyMaze(FrameBuffer[drawBuffer], MazeBuf, SCREEN_WIDTH * SCREEN_HEIGHT * 4 / 256);

    checkAndLoadTGA(&aImage[kImgLevelSelect], 1, TRUE);
    SPRBlitter(aImage[kImgLevelSelect].spr.img, aImage[kImgLevelSelect].spr.dmaflags, 0, 0,360,32,
	       FrameBuffer[drawBuffer], DMAFLAGS, SCREEN_WIDTH/2, SCREEN_HEIGHT - 22,
	       0x14000 * SCREEN_WIDTH / 720, 0x14000 * SCREEN_HEIGHT / 480,
	       0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);

    levelNum = curSelection + 1;
    x = 205;
    y = SCREEN_HEIGHT - 22;
    offsetX = (levelNum % 10) & 0x7;
    offsetY = 4 + ((levelNum % 10) >> 3);
    SPRBlitter(aImage[kImgFont].spr.img, aImage[kImgFont].spr.dmaflags,
               1 + (offsetX*32), 1 + (offsetY*32), 30, 30,
	       FrameBuffer[drawBuffer], DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0,
	       0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, aImage[kImgFont].transColor);

    x -= 20 * SCORESCALE / 0x10000;
    if (levelNum > 9) {
      offsetX = ((levelNum / 10) % 10) & 0x7;
      offsetY = 4 + (((levelNum / 10) % 10) >> 3);
      SPRBlitter(aImage[kImgFont].spr.img, aImage[kImgFont].spr.dmaflags,
                 1 + (offsetX*32), 1 + (offsetY*32), 30, 30,
		 FrameBuffer[drawBuffer], DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0,
		 0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, aImage[kImgFont].transColor);
    }
    
    x -= 20 * SCORESCALE / 0x10000;
    if (levelNum > 99) {
      offsetX = ((levelNum / 100) % 10) & 0x7;
      offsetY = 4 + (((levelNum / 100) % 10) >> 3);
      SPRBlitter(aImage[kImgFont].spr.img, aImage[kImgFont].spr.dmaflags,
                 1 + (offsetX*32), 1 + (offsetY*32), 30, 30,
		 FrameBuffer[drawBuffer], DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0,
		 0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, aImage[kImgFont].transColor);
    }

    if (blink) {
      myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, 210, SCREEN_HEIGHT - 30, 16, 16, kBlack);
      myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, 118, SCREEN_HEIGHT - 30, 16, 16, kBlack);
    }

    blink = 1 - blink;
    
    // logical OR between joystick and remote
    prevJoystick = joystick;
    joystick = _Controller[1].buttons | _Controller[0].buttons;

    if (_Controller[1].d2.yAxis >  DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_UP;
    if (_Controller[1].d2.yAxis < -DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_DOWN;
    if (_Controller[1].d1.xAxis >  DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_RIGHT;
    if (_Controller[1].d1.xAxis < -DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_LEFT;
    
    if ((joystick & CTRLR_DPAD_UP) && !(prevJoystick & CTRLR_DPAD_UP)) {
      myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection += 10;
      newMaze = randomize = 1;
    } else if ((joystick & CTRLR_DPAD_DOWN) && !(prevJoystick & CTRLR_DPAD_DOWN)) {
      myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection -= 10;
      newMaze = randomize = 1;
    } else if ((joystick & CTRLR_DPAD_LEFT) && !(prevJoystick & CTRLR_DPAD_LEFT)) {
      myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection--;
      newMaze = randomize = 1;
    } else if ((joystick & CTRLR_DPAD_RIGHT) && !(prevJoystick & CTRLR_DPAD_RIGHT)) {
      myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection++;
      newMaze = randomize = 1;
    }

    if (curSelection < 0) curSelection = maxLevels - 1;
    if (curSelection >= maxLevels) curSelection = 0;

    if (joystick & JOY_A) done = 1;
    if (joystick & JOY_B) { done = 1; curSelection = -1; }

    if ((joystick & JOY_C_UP   ) && !(prevJoystick & JOY_C_UP   ))
      { wallIndex  = (wallIndex  + 1) % wallMax;   newMaze = 1; }
    if ((joystick & JOY_C_DOWN ) && !(prevJoystick & JOY_C_DOWN ))
      { pillIndex  = (pillIndex  + 1) % pillMax;   newMaze = 1; }
    if ((joystick & JOY_C_LEFT ) && !(prevJoystick & JOY_C_LEFT ))
      { fruitIndex = (fruitIndex + 1) % fruitMax;  newMaze = 1; }
    if ((joystick & JOY_C_RIGHT) && !(prevJoystick & JOY_C_RIGHT))
      { ghostIndex = (ghostIndex + 1) % ghostMax;  newMaze = 1; }

    // Process display list
    SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH - 1,
		     SCREEN_HEIGHT - 1, kBlack);
    SPRDraw(0,0);
    SPRWait();

    if (!done && (gameMode == MODE_TOURNAMENT) && (highScoreTimeout > 0) && !newMaze) {
      highScoreTimeout--;
      x = (SCREEN_WIDTH  / 2) - (10 * 24 * SCORESCALE) / (0x10000 * 2);
      y = (SCREEN_HEIGHT / 2) - 50;
      drawString("high score\0", FrameBuffer[drawBuffer], x, y, 0x40000000);
      y += 2 * (24 * SCORESCALE) / 0x10000;
      x -= 2 * (24 * SCORESCALE) / 0x10000;
      if (aLevelScore[curSelection].name[0]) {
	drawString(aLevelScore[curSelection].name, FrameBuffer[drawBuffer], x, y, 0x40000000);
      } else {
	drawString("nobody\0", FrameBuffer[drawBuffer], x, y, 0x40000000);
      }
      x += 9 * (24 * SCORESCALE) / 0x10000;
      percent = aLevelScore[curSelection].percent;
      drawScore(aLevelScore[curSelection].score, FrameBuffer[drawBuffer], x, y, 0x40000000, FALSE);

      if (aLevelScore[curSelection].name[0]) {
	x = (SCREEN_WIDTH  / 2) - (15 * 24 * SCORESCALE) / (0x10000 * 2);
	y += 2 * (24 * SCORESCALE) / 0x10000;
	drawString("code\0", FrameBuffer[drawBuffer], x, y, 0x40000000);
	x += 5 * (24 * SCORESCALE) / 0x10000;
        verfCode = getVerfCode(aLevelScore[curSelection].score, aLevelScore[curSelection].name,
                               curSelection+1);
        if (emulator) verfCode = 0;

	for (i = 0; i < 11; i++) {
	  tmpString[10-i] = ((verfCode >> (3*i)) & 0x7) + '0';
	  if (i == 10) tmpString[10-i] += 4;  // Force top bit of most-significant digit to be one
	}
	tmpString[11] = 0;
	drawString(tmpString, FrameBuffer[drawBuffer], x, y, 0x40000000);
      }
    }

    SPRWait();

    VideoSync();
  }

  // Need to delete and recreate ghosts since sprite changes every level
  for(i = 0; i < MAX_GHOSTS; i++) {
    if (ghosts[i].spr) {
      SPRRemoveSprite(ghosts[i].spr);
      SPRDeleteSprite(ghosts[i].spr);
      ghosts[i].spr = NULL;
    }
  }

  SPRRemoveSprite(chomper.chomperman);

  return(curSelection);
}

//==================================================================================================
//==================================================================================================

void initRandomArray()
{
  int i;
  srand(VERF_SEED);
  for (i = 0; i < RAND_ARRAY_ENTRIES; i++) aRand[i] = rand();
  //  printf("\n");
  //  for (i = 0; i < RAND_ARRAY_ENTRIES; i++) printf("%08X  ", aRand[i]);
  //  printf("\n");
}

//==================================================================================================
//==================================================================================================

void getName(char *name)
{
  int i;
  int done = 0;
  int x, y;
  int joystick = 0, prevJoystick;
  int offsetX, offsetY;
  int curSelection = 0;
  int charIndex = 0;
  char *ptr;
  int blink = 0;

#define SEL_BACKSP 42
#define SEL_END    43

  while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));

  if (name[0] == 0) { name[0] = 'A';  name[1] = 0; }
  else { charIndex = strlen(name); curSelection = SEL_END; }

  while (!done) {
    myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);

    checkAndLoadTGA(&aImage[kImgNameEntry], 1, TRUE);
    SPRBlitter(aImage[kImgNameEntry].spr.img, aImage[kImgNameEntry].spr.dmaflags, 0, 0, 360, 160,
	       FrameBuffer[drawBuffer], DMAFLAGS, SCREEN_WIDTH/2, SCREEN_HEIGHT/2,
	       0x0E000, 0x0E000, 0, 0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, 0);

    // logical OR between joystick and remote
    prevJoystick = joystick;
    joystick = _Controller[1].buttons | _Controller[0].buttons;

    if (_Controller[1].d2.yAxis >  DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_UP;
    if (_Controller[1].d2.yAxis < -DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_DOWN;

    if ((joystick & CTRLR_DPAD_UP) && !(prevJoystick & CTRLR_DPAD_UP)) {
      myPlayWAV((void *) ChomperChompWav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection++;
    }

    if ((joystick & CTRLR_DPAD_DOWN) && !(prevJoystick & CTRLR_DPAD_DOWN)) {
      myPlayWAV((void *) ChomperChompWav, -1, &Pan, 0x40000000, 0x40000000);
      curSelection--;
    }

    // Wrap selection around
    #define MAX_SELECTION (SEL_END+1)
    if (curSelection <              0) curSelection += MAX_SELECTION;
    if (curSelection >= MAX_SELECTION) curSelection -= MAX_SELECTION;

    if (charIndex == 8) {
      // If have all 8 characters only display BACKSPACE and END
      if (curSelection < SEL_BACKSP) curSelection = SEL_END;
      if (curSelection > SEL_END   ) curSelection = SEL_BACKSP;
    } else if (charIndex == 0) {
      // If have no characters yet, do not display BACKSPACE or END
      if (curSelection == SEL_BACKSP) curSelection = 0;
      if (curSelection == SEL_END   ) curSelection = SEL_BACKSP-1;
    }

    // Translate selection to a character
    if (curSelection < 26)
      name[charIndex] = 'A' + (curSelection - 0);
    else if (curSelection < (26 + 10))
      name[charIndex] = '0' + (curSelection - 26);
    else {
      switch (curSelection) {
        case 36:  name[charIndex] = '-';  break;
        case 37:  name[charIndex] = '\''; break;
        case 38:  name[charIndex] = '_';  break;
        case 39:  name[charIndex] = '!';  break;
        case 40:  name[charIndex] = '?';  break;
        case 41:  name[charIndex] = '.';  break;
        default:  name[charIndex] = '*';  break;
      }
    }
    name[charIndex + 1] = 0;

    // Check if user hit button to accept/delete
    if ((joystick & JOY_A) || (joystick & JOY_B)) {
      myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
      while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));

      if ((joystick & JOY_B) || (curSelection == SEL_BACKSP)) {

	// Player hit B or the backspace character
	if (charIndex > 0) {
	  name[charIndex] = 0;
	  charIndex--;
	  if ((name[charIndex] >= 'A') && (name[charIndex] <= 'Z'))
	    curSelection = name[charIndex] - 'A';
	  else if ((name[charIndex] >= '0') && (name[charIndex] <= '9'))
	    curSelection = name[charIndex] - '9';
	  else {
	    switch (name[charIndex]) {
              case '-':  curSelection = 36;  break;
              case '\'': curSelection = 37;  break;
              case '_':  curSelection = 38;  break;
              case '!':  curSelection = 39;  break;
              case '?':  curSelection = 40;  break;
              case '.':  curSelection = 41;  break;
	    }
	  }
	}

      } else if ((joystick & JOY_A) && (curSelection == SEL_END)) {
	// Player hit END
	name[charIndex] = 0;
	done = 1;

      } else if (joystick & JOY_A) {
	// Player hit A to accept a character
	if (charIndex < 8) {
	  charIndex++;
	  name[charIndex] = 0;
	  if (curSelection != SEL_BACKSP) curSelection = 0;
	}
      }
    }
    
    x = 185;
    y = 143;
    for (ptr = name, i = 0; *ptr; ptr++, i++) {
      if ((*ptr >= 'A') && (*ptr <= 'Z')) {
	offsetX = (*ptr - 'A') & 0x7;
	offsetY = (*ptr - 'A') >> 3;
      } else if ((*ptr >= '0') && (*ptr <= '9')) {
	offsetX = (*ptr - '0') & 0x7;
	offsetY = 4 + ((*ptr - '0') >> 3);
      } else if ((i == charIndex) && (curSelection >= SEL_BACKSP)) {
	switch (curSelection) {
	  case SEL_BACKSP:  offsetX = 5; offsetY = 5; break;
	  case SEL_END:     offsetX = 4; offsetY = 5; break;
          default:          offsetX = 2; offsetY = 5; printf("%d", curSelection); fflush(stdout); break;
	}
      } else {
	switch (*ptr) {
	  case '-':  offsetX = 3; offsetY = 3; break;
	  case '\'': offsetX = 4; offsetY = 3; break;
	  case '_':  offsetX = 5; offsetY = 3; break;
	  case '!':  offsetX = 6; offsetY = 3; break;
	  case '?':  offsetX = 7; offsetY = 3; break;
	  case '.':  offsetX = 3; offsetY = 5; break;
          default:   offsetX = 2; offsetY = 5; printf("%d", curSelection); fflush(stdout); break;
	}
      }
      if ((i != charIndex) || blink)
	SPRBlitter(aImage[kImgFont].spr.img, aImage[kImgFont].spr.dmaflags,
                   1 + (offsetX*32), 1 + (offsetY*32),30,30,
		   FrameBuffer[drawBuffer], DMAFLAGS, x, y, 0x0B000, 0x0B000, 0,
		   0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, aImage[kImgFont].transColor);
      x += (24 * 0x0B000) / 0x10000;
    }
    blink = 1 - blink;

    // Process display list
    //    SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH - 1,
    //		     SCREEN_HEIGHT - 1, kBlack);
    //    SPRDraw(0,0);
    SPRWait();
    VideoSync();
  }

  while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));
}
      
//==================================================================================================
//==================================================================================================

void showVerfCode(int level)
{
  int i;
  int done = 0;
  int x, y;
  int joystick = 0, prevJoystick;
  int verfCode;
  char tmpString[12];

  while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));

  verfCode = getVerfCode(aLevelScore[level].score, aLevelScore[level].name, level+1);
  if (emulator) verfCode = 0;

  while (!done) {
    myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);

    checkAndLoadTGA(&aImage[kImgVerfCode], 1, TRUE);
    SPRBlitter(aImage[kImgVerfCode].spr.img, aImage[kImgVerfCode].spr.dmaflags, 0, 0, 360, 240,
	       FrameBuffer[drawBuffer], DMAFLAGS, SCREEN_WIDTH/2, SCREEN_HEIGHT/2,
	       0x0E000, 0x0E000, 0, 0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, 0);

    x = 200;
    y = 132;
    drawString(aLevelScore[level].name, FrameBuffer[drawBuffer], x, y, 0x40000000);
    y += 18;
    percent = 0;
    sprintf(tmpString, "%d", aLevelScore[level].score);
    drawString(tmpString, FrameBuffer[drawBuffer], x, y, 0x40000000);
    y += 18;
    sprintf(tmpString, "%d", level+1);
    drawString(tmpString, FrameBuffer[drawBuffer], x, y, 0x40000000);
    y += 18;
    drawString(versionString, FrameBuffer[drawBuffer], x, y, 0x40000000);
    y += 18;
    for (i = 0; i < 11; i++) {
      tmpString[10-i] = ((verfCode >> (3*i)) & 0x7) + '0';
      if (!emulator && (i == 10))
        tmpString[10-i] += 4;  // Force top bit of most-significant digit to be one
    }
    tmpString[11] = 0;
    drawString(tmpString, FrameBuffer[drawBuffer], x, y, 0x40000000);
    

    // logical OR between joystick and remote
    prevJoystick = joystick;
    joystick = _Controller[1].buttons | _Controller[0].buttons;

    if ((joystick & JOY_A) || (joystick & JOY_B)) done = 1;
    
    // Process display list
    //    SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH - 1,
    //		     SCREEN_HEIGHT - 1, kBlack);
    //    SPRDraw(0,0);
    SPRWait();
    VideoSync();
  }

  myPlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);

  while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));
}
      
//==================================================================================================
//==================================================================================================

void printMaze(char *maze, int w)
{
  int rowCount;

  for (rowCount = 1; *maze; maze++, rowCount++) {
    printf("%c", *maze);
    if (rowCount == w) { printf("\n");  rowCount = 0; }
  }

  fflush(stdout);
}

//==================================================================================================
//==================================================================================================

void myMemCpy(char *dst, const char *src, int numChar)
{
  char *tmpBuf;

  tmpBuf = (char *) malloc (numChar * sizeof(char));
  if (tmpBuf) {
    memcpy(tmpBuf, src, numChar);
    memcpy(dst, tmpBuf, numChar);
  } else {
    printf("\nERROR: Could not allocate %d bytes for tmpBuf in myMemCpy()", numChar);
    fflush(stdout);
  }
  free(tmpBuf);
}

//==================================================================================================
//==================================================================================================

int editLevel(int levelIndex)
{
  int returnValue = -1;
  int done = FALSE;
  char *ptr, tmpChar;
  LEVEL *curLevel = &aLevelEdited[levelIndex];
  int newMaze = TRUE;
  int joystick = 0, prevJoystick = -1, joyEdge;
  char *mazeChar;
  unsigned long wallScale = 0, pillScale = 0, fruitScale = 0;
  //  SPRITE *cursor[4];  // Do not know why the array does not work
  SPRITE *cursor0, *cursor1, *cursor2, *cursor3;

  // Editor state
  int drag = 0;
  int cursorX = 0x00008000, cursorY = 0x00008000, prevX, prevY;
  int mirrorHoriz = TRUE, mirrorVert = TRUE;
  int overwrite = FALSE;
  int optionMenu = FALSE;
  int optionIndex = 0;
  int optionChoose = FALSE;
  int newValue = 0;
  int showChars = FALSE;
  int dpadDelay = 0;

  // Temporary variables
  int i, x, y;

  // Remove Chomperman sprite
  SPRRemoveSprite(chomper.chomperman);

  // Make sure enemy graphic set is loaded
  checkAndLoadTGA(&aGhostImg[ghostIndex], 7, TRUE);
  checkAndLoadTGA(&aImage[kImgLevelEditOther], 1, TRUE);

  EMU_PRINT("\n\nWelcome to the level editor");
  EMU_PRINT("\nTo print your level, press and hold the NUON button.");
  EMU_PRINT("\nThen press and hold the L & R buttons.\n");
  
  // Initialize enemy sprites
  for (i = 0; i < MAX_GHOSTS; i++) {
    if (ghosts[i].spr) {
      SPRRemoveSprite(ghosts[i].spr);
      SPRDeleteSprite(ghosts[i].spr);
    }
    ghosts[i].spr = SPRCreateSprite(&aGhostImg[ghostIndex].spr,
				    i * aGhostImg[ghostIndex].w, 0,
				    aGhostImg[ghostIndex].w, aGhostImg[ghostIndex].w);
  }

  // Initialize first time through
  if ((strlen(curLevel->widthCalib) == 0) || (strlen(curLevel->maze) == 0)) {
    for (i = 0, ptr = curLevel->widthCalib; i < MAZE_MAX_W; i++, ptr++) *ptr = ' ';
    *ptr = 0;
    for (i = 0, ptr = curLevel->maze; i < (MAZE_MAX_W * MAZE_MAX_H); i++, ptr++) *ptr = ' ';
    *ptr = 0;
    curLevel->w = MAZE_MAX_W;
    curLevel->h = MAZE_MAX_H;
    curLevel->offsetX = 0;
    curLevel->offsetY = 0;
  }

  cursor0 = SPRCreateSprite(&aImage[kImgMisc].spr, 32, 64, 32, 32);
  cursor1 = SPRCreateSprite(&aImage[kImgMisc].spr, 32, 64, 32, 32);
  cursor2 = SPRCreateSprite(&aImage[kImgMisc].spr, 32, 64, 32, 32);
  cursor3 = SPRCreateSprite(&aImage[kImgMisc].spr, 32, 64, 32, 32);
  SPRAddSprite(cursor0, SPRITE_SCREEN_X(cursorX), SPRITE_SCREEN_Y(cursorY), 0, 0x10000, 0x10000,
	       kSpriteInterpolatedTrans | (1<<30), 0, 2);
 
  memcpy(maze, curLevel->maze, curLevel->w * curLevel->h);

  while (!done) {

#define DPAD_DELAY  10

#define DRAG_WALL_ON   1
#define DRAG_WALL_OFF  2
#define DRAG_PILL_ON   3
#define DRAG_PILL_OFF  4
#define DRAG_FRUIT_ON  5
#define DRAG_FRUIT_OFF 6
#define DRAG_ERASE     7
#define DRAG_POWERPILL_ON  8
#define DRAG_POWERPILL_OFF 9
#define DRAG_SLOW_ON  10
#define DRAG_SLOW_OFF 11

    //
    // User Input
    //
    if (prevJoystick != -1) prevJoystick = joystick;
    joystick = _Controller[1].buttons | _Controller[0].buttons;
    if (prevJoystick == -1) prevJoystick = joystick;

    if (_Controller[1].d2.yAxis >  DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_UP;
    if (_Controller[1].d2.yAxis < -DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_DOWN;
    if (_Controller[1].d1.xAxis >  DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_RIGHT;
    if (_Controller[1].d1.xAxis < -DEADZONE_CENTER)
      joystick |= CTRLR_DPAD_LEFT;

    joyEdge = (joystick ^ prevJoystick) & joystick;
    if ((joystick ^ prevJoystick) & ~(JOY_UP | JOY_DOWN | JOY_LEFT | JOY_RIGHT)) drag = 0;

    prevX = cursorX;
    prevY = cursorY;

    if (!optionMenu) {
      if (dpadDelay) dpadDelay--;
      if (!(joystick & (JOY_UP | JOY_DOWN | JOY_LEFT | JOY_RIGHT))) dpadDelay = 0;

      if (_Controller[1].d1.xAxis >  DEADZONE_CENTER) {
	cursorX += (_Controller[1].d1.xAxis - DEADZONE_CENTER) << 8;

      } else if (_Controller[1].d1.xAxis < -DEADZONE_CENTER) {
	cursorX -= (-_Controller[1].d1.xAxis - DEADZONE_CENTER) << 8;

      } else if (((dpadDelay == 1) && (joystick & JOY_RIGHT)) ||
		 (!dpadDelay       && (joyEdge  & JOY_RIGHT))) {
	cursorX += 1<<16;
	dpadDelay = DPAD_DELAY;

      } else if (((dpadDelay == 1) && (joystick & JOY_LEFT)) ||
		 (!dpadDelay       && (joyEdge  & JOY_LEFT))) {
	cursorX -= 1<<16;
	dpadDelay = DPAD_DELAY;

      } else {
	cursorX = (cursorX & 0xFFFF0000) | 0x8000;
      }

      if (cursorX <  0) cursorX += curLevel->w << 16;
      if (cursorX >= (curLevel->w<<16)) cursorX -= curLevel->w << 16;


      if (_Controller[1].d2.yAxis >  DEADZONE_CENTER) {
	cursorY -= (_Controller[1].d2.yAxis - DEADZONE_CENTER) << 8;

      } else if (_Controller[1].d2.yAxis < -DEADZONE_CENTER) {
	cursorY += (-_Controller[1].d2.yAxis - DEADZONE_CENTER) << 8;

      } else if ((((dpadDelay == 1) ||(dpadDelay == DPAD_DELAY)) && (joystick & JOY_DOWN)) ||
		 (!dpadDelay       && (joyEdge  & JOY_DOWN))) {
	cursorY += 1<<16;
	dpadDelay = DPAD_DELAY;

      } else if ((((dpadDelay == 1) ||(dpadDelay == DPAD_DELAY)) && (joystick & JOY_UP)) ||
		 (!dpadDelay       && (joyEdge  & JOY_UP))) {
	cursorY -= 1<<16;
	dpadDelay = DPAD_DELAY;

      } else {
	cursorY = (cursorY & 0xFFFF0000) | 0x8000;
      }

      if (cursorY <  0) cursorY += curLevel->h << 16;
      if (cursorY >= (curLevel->h<<16)) cursorY -= curLevel->h << 16;
    }


#define OPTION_CLEAR         0
#define OPTION_MAZE_W        1
#define OPTION_MAZE_H        2
#define OPTION_MIRROR_HORIZ  3
#define OPTION_MIRROR_VERT   4
#define OPTION_OVERWRITE     5
#define OPTION_REMOVE        6
#define OPTION_INSERT        7
#define OPTION_SHIFT         8
#define OPTION_TEST          9
#define OPTION_QUIT          10

#define MAX_OPTIONS          11

    // Check for button
    if (optionMenu) {
      if (joyEdge & JOY_A) {
	switch (optionIndex) {
	  case OPTION_QUIT:  returnValue = -1;  done = TRUE;  break;
	  case OPTION_TEST:
	    for (ptr = maze; *ptr && (*ptr != 'X'); ptr++);
	    if (*ptr == 'X') { returnValue = 0; newMaze = TRUE; done = TRUE; optionMenu = FALSE; }
	    break;

  	  case OPTION_MIRROR_HORIZ:  mirrorHoriz = 1 - mirrorHoriz;  break;
  	  case OPTION_MIRROR_VERT:   mirrorVert  = 1 - mirrorVert;   break;
	  case OPTION_OVERWRITE:     overwrite   = 1 - overwrite;    break;

 	  case OPTION_CLEAR:
	    if (optionChoose) {
	      memset(maze, ' ', curLevel->h * curLevel->w);
	      newMaze = TRUE;
	      optionChoose = FALSE;
	    } else
	      optionChoose = TRUE;
	    break;

 	  case OPTION_MAZE_W:
	    if (!optionChoose) {
	      optionChoose = TRUE;
	      newValue = curLevel->w;
	    } else {
	      optionChoose = FALSE;

	      // Perform different actions if we are adding or removing cols
	      if (newValue > curLevel->w) {
		// Shift all rows to make room for new tiles
		ptr = &maze[0];
		for (i = (curLevel->h - 1); i > 0; i--, ptr += newValue) {
		  myMemCpy(ptr + newValue, ptr + curLevel->w, i * curLevel->w);
		  memset(ptr + curLevel->w, ' ', newValue - curLevel->w);
		}
		memset(ptr + curLevel->w, ' ', newValue - curLevel->w);

		newMaze = TRUE;
		curLevel->w = newValue;
		maze[curLevel->w * curLevel->h] = 0;

	      } else if (newValue < curLevel->w) {
		ptr = &maze[0];
		// Shift all rows to elminate unused tiles
		for (i = (curLevel->h - 1); i > 0; i--, ptr += newValue)
		  myMemCpy(ptr + newValue, ptr + curLevel->w, i * curLevel->w);

		newMaze = TRUE;
		curLevel->w = newValue;
		maze[curLevel->w * curLevel->h] = 0;
	      }
	    }
	    break;

 	  case OPTION_MAZE_H:
	    if (!optionChoose) {
	      optionChoose = TRUE;
	      newValue = curLevel->h;
	    } else {
	      optionChoose = FALSE;

	      // Perform different actions if we are adding or removing rows
	      if (newValue > curLevel->h) {
		ptr = &maze[curLevel->w * curLevel->h];
		// Simply clear the rows after the last one
		for (i = 0; i < (newValue - curLevel->h) * curLevel->w; i++, ptr++) *ptr = ' ';

		newMaze = TRUE;
		curLevel->h = newValue;
		maze[curLevel->w * curLevel->h] = 0;

	      } else if (newValue < curLevel->h) {
		// Just truncate the end of the string to eliminate rows
		newMaze = TRUE;
		curLevel->h = newValue;
		maze[curLevel->w * curLevel->h] = 0;
	      }
	    }
	    break;

  	  case OPTION_REMOVE:
  	  case OPTION_INSERT:
  	  case OPTION_SHIFT:
	    optionChoose = optionChoose ? FALSE : TRUE;
	    break;
	}
      }

      if (optionChoose && (joyEdge & JOY_RIGHT)) {
	switch (optionIndex) {
	  case OPTION_MAZE_W:  if (newValue < MAZE_MAX_W) newValue++;  break;
	  case OPTION_MAZE_H:  if (newValue < MAZE_MAX_H) newValue++;  break;

 	  case OPTION_SHIFT:
	    ptr = &maze[0];
	    for (i = 0; i < curLevel->h; i++, ptr += curLevel->w) {
	      tmpChar = ptr[curLevel->w - 1];
	      myMemCpy(ptr+1, ptr, curLevel->w-1);
	      ptr[0] = tmpChar;
	    }
	    newMaze = TRUE;
	    break;

	  case OPTION_INSERT:
	    // Insert a column
	    x = cursorX>>16;
	    if (mirrorHoriz && ((x < (curLevel->w>>1)) ||
				((x == (curLevel->w>>1)) && (curLevel->w & 0x1)))) {
	      ptr = &maze[0];
	      for (i = 0; i < curLevel->h; i++, ptr += curLevel->w)
		myMemCpy(ptr, ptr+1, x);
	    } else {
	      ptr = &maze[x];
	      for (i = 0; i < curLevel->h; i++, ptr += curLevel->w)
		myMemCpy(ptr+1, ptr, curLevel->w - x - 1);
	    }

	    // Remove the mirror row if it exists and isn't the middle row
	    if (mirrorHoriz) {
	      x = (curLevel->w-1) - x;
	      if (x < (curLevel->w>>1)) {
		ptr = &maze[0];
		for (i = 0; i < curLevel->h; i++, ptr += curLevel->w)
		  myMemCpy(ptr, ptr+1, x);
	      } else {
		ptr = &maze[x];
		for (i = 0; i < curLevel->h; i++, ptr += curLevel->w)
		  myMemCpy(ptr+1, ptr, curLevel->w - x - 1);
	      }

	      for (i = 0, ptr = &maze[x]; i < curLevel->h; i++, ptr += curLevel->w) *ptr = ' ';
	    }

	    // Clear column with cursor
	    x = cursorX>>16;
	    for (i = 0, ptr = &maze[x]; i < curLevel->h; i++, ptr += curLevel->w) *ptr = ' ';

	    newMaze = TRUE;
	    break;

	  case OPTION_REMOVE:
	    // Remove the current col
	    x = cursorX>>16;
	    if (!mirrorHoriz || (x < (curLevel->w>>1))) {
	      ptr = &maze[0];
	      for (i = 0; i < curLevel->h; i++, ptr += curLevel->w) {
		myMemCpy(ptr+1, ptr, x);
		ptr[0] = ' ';
	      }
	    } else {
	      ptr = &maze[x];
	      for (i = 0; i < curLevel->h; i++, ptr += curLevel->w) {
		myMemCpy(ptr, ptr+1, curLevel->w - x - 1);
		ptr[curLevel->w-x-1] = ' ';
	      }
	    }

	    // Remove the mirror row if it exists and isn't the middle row
	    if (mirrorHoriz) {
	      x = (curLevel->w-1) - x;
	      if ((x < (curLevel->w>>1)) ||
		  ((x == (curLevel->w>>1)) && (curLevel->w & 0x1))) {
		ptr = &maze[0];
		for (i = 0; i < curLevel->h; i++, ptr += curLevel->w) {
		  myMemCpy(ptr+1, ptr, x);
		  ptr[0] = ' ';
		}
	      } else {
		ptr = &maze[x];
		for (i = 0; i < curLevel->h; i++, ptr += curLevel->w) {
		  myMemCpy(ptr, ptr+1, curLevel->w - x - 1);
		  ptr[curLevel->w-x-1] = ' ';
		}
	      }
	    }

	    newMaze = TRUE;
	    break;
	}
      }

      if (optionChoose && (joyEdge & JOY_LEFT)) {
	switch (optionIndex) {
	  case OPTION_MAZE_W:  if (newValue > 4) newValue--;  break;
	  case OPTION_MAZE_H:  if (newValue > 4) newValue--;  break;

 	  case OPTION_SHIFT:
	    ptr = &maze[0];
	    for (i = 0; i < curLevel->w; i++, ptr += curLevel->w) {
	      tmpChar = ptr[0];
	      myMemCpy(ptr, ptr+1, curLevel->w-1);
	      ptr[curLevel->w - 1] = tmpChar;
	    }
	    newMaze = TRUE;
	    break;

	  case OPTION_INSERT:
	    // Insert a column
	    x = cursorX>>16;
	    if (!mirrorHoriz || (x < (curLevel->w>>1))) {
	      ptr = &maze[0];
	      for (i = 0; i < curLevel->h; i++, ptr += curLevel->w)
		myMemCpy(ptr, ptr+1, x);
	    } else {
	      ptr = &maze[x];
	      for (i = 0; i < curLevel->h; i++, ptr += curLevel->w)
		myMemCpy(ptr+1, ptr, curLevel->w - x - 1);
	    }

	    // Remove the mirror row if it exists and isn't the middle row
	    if (mirrorHoriz) {
	      x = (curLevel->w-1) - x;
	      if ((x < (curLevel->w>>1)) ||
		  ((x == (curLevel->w>>1)) && (curLevel->w & 0x1))) {
		ptr = &maze[0];
		for (i = 0; i < curLevel->h; i++, ptr += curLevel->w)
		  myMemCpy(ptr, ptr+1, x);
	      } else {
		ptr = &maze[x];
		for (i = 0; i < curLevel->h; i++, ptr += curLevel->w)
		  myMemCpy(ptr+1, ptr, curLevel->w - x - 1);
	      }

	      for (i = 0, ptr = &maze[x]; i < curLevel->h; i++, ptr += curLevel->w) *ptr = ' ';
	    }

	    // Clear column with cursor
	    x = cursorX>>16;
	    for (i = 0, ptr = &maze[x]; i < curLevel->h; i++, ptr += curLevel->w) *ptr = ' ';

	    newMaze = TRUE;
	    break;

	  case OPTION_REMOVE:
	    // Remove the current col
	    x = cursorX>>16;
	    if (mirrorHoriz && ((x < (curLevel->w>>1)) ||
				((x == (curLevel->w>>1)) && (curLevel->w & 0x1)))) {
	      ptr = &maze[0];
	      for (i = 0; i < curLevel->h; i++, ptr += curLevel->w) {
		myMemCpy(ptr+1, ptr, x);
		ptr[0] = ' ';
	      }
	    } else {
	      ptr = &maze[x];
	      for (i = 0; i < curLevel->h; i++, ptr += curLevel->w) {
		myMemCpy(ptr, ptr+1, curLevel->w - x - 1);
		ptr[curLevel->w-x-1] = ' ';
	      }
	    }

	    // Remove the mirror row if it exists and isn't the middle row
	    if (mirrorHoriz) {
	      x = (curLevel->w-1) - x;
	      if (x < (curLevel->w>>1)) {
		ptr = &maze[0];
		for (i = 0; i < curLevel->h; i++, ptr += curLevel->w) {
		  myMemCpy(ptr+1, ptr, x);
		  ptr[0] = ' ';
		}
	      } else {
		ptr = &maze[x];
		for (i = 0; i < curLevel->h; i++, ptr += curLevel->w) {
		  myMemCpy(ptr, ptr+1, curLevel->w - x - 1);
		  ptr[curLevel->w-x-1] = ' ';
		}
	      }
	    }

	    newMaze = TRUE;
	    break;
	}
      }

      if (optionChoose && (joyEdge & JOY_UP)) {
	switch (optionIndex) {
 	  case OPTION_SHIFT:
	    for (i = 0; i < curLevel->w; i++) {
	      tmpChar = maze[0];
	      myMemCpy(&maze[0], &maze[1], (curLevel->w * curLevel->h) - 1);
	      maze[(curLevel->w * curLevel->h) - 1] = tmpChar;
	    }
	    newMaze = TRUE;
	    break;

	  case OPTION_INSERT:
	    // Insert a new row
	    y = cursorY>>16;
	    if (!mirrorVert || (y < (curLevel->h>>1))) {
	      ptr = &maze[0];
	      for (i = 0; i < y; i++, ptr += curLevel->w)
		memcpy(ptr, ptr + curLevel->w, curLevel->w);
	    } else {
	      ptr = &maze[curLevel->w * (curLevel->h - 1)];
	      for (i = curLevel->h - 1; i > y; i--, ptr -= curLevel->w)
		memcpy(ptr, ptr - curLevel->w, curLevel->w);
	    }

	    // Remove the mirror row if it exists and isn't the middle row
	    if (mirrorVert) {
	      y = (curLevel->h-1) - y;
	      if ((y < (curLevel->h>>1)) ||
		  ((y == (curLevel->h>>1)) && (curLevel->h & 0x1))) {
		ptr = &maze[0];
		for (i = 0; i < y; i++, ptr += curLevel->w)
		  memcpy(ptr, ptr + curLevel->w, curLevel->w);
	      } else {
		ptr = &maze[curLevel->w * (curLevel->h - 1)];
		for (i = curLevel->h - 1; i > y; i--, ptr -= curLevel->w)
		  memcpy(ptr, ptr - curLevel->w, curLevel->w);
	      }

	      memset(&maze[y*curLevel->w], ' ', curLevel->w);
	    }

	    y = cursorY>>16;
	    memset(&maze[y*curLevel->w], ' ', curLevel->w);

	    newMaze = TRUE;
	    break;


	  case OPTION_REMOVE:
	    // Remove the current row
	    y = cursorY>>16;
	    ptr = &maze[y*curLevel->w];
	    if (mirrorVert && ((y < (curLevel->h>>1)) ||
			       ((y == (curLevel->h>>1)) && (curLevel->h & 0x1)))) {
	      for (i = y; i > 0; i--, ptr -= curLevel->w)
		memcpy(ptr, ptr - curLevel->w, curLevel->w);
	    } else {
	      for (i = y; i < (curLevel->h-1); i++, ptr += curLevel->w)
		memcpy(ptr, ptr + curLevel->w, curLevel->w);
	    }
	    for (i = 0; i < curLevel->w; i++, ptr++) *ptr = ' ';

	    // Remove the mirror row
	    if (mirrorVert) {
	      y = (curLevel->h-1) - y;
	      ptr = &maze[y*curLevel->w];
	      if (y < (curLevel->h>>1)) {
		for (i = y; i > 0; i--, ptr -= curLevel->w)
		  memcpy(ptr, ptr - curLevel->w, curLevel->w);
	      } else {
		for (i = y; i < (curLevel->h-1); i++, ptr += curLevel->w)
		  memcpy(ptr, ptr + curLevel->w, curLevel->w);
	      }
	      for (i = 0; i < curLevel->w; i++, ptr++) *ptr = ' ';
	    }

	    newMaze = TRUE;
	    break;
	}
      }

      if (optionChoose && (joyEdge & JOY_DOWN)) {
	switch (optionIndex) {
 	  case OPTION_SHIFT:
	    for (i = 0; i < curLevel->w; i++) {
	      tmpChar = maze[(curLevel->w * curLevel->h) - 1];
	      myMemCpy(&maze[1], &maze[0], (curLevel->w * curLevel->h) - 1);
	      maze[0] = tmpChar;
	    }
	    newMaze = TRUE;
	    break;

	  case OPTION_INSERT:
	    // Insert a new row
	    y = cursorY>>16;
	    if (mirrorVert && ((y < (curLevel->h>>1))  ||
			       ((y == (curLevel->h>>1)) && (curLevel->h & 0x1)))) {
	      ptr = &maze[0];
	      for (i = 0; i < y; i++, ptr += curLevel->w)
		memcpy(ptr, ptr + curLevel->w, curLevel->w);
	    } else {
	      ptr = &maze[curLevel->w * (curLevel->h - 1)];
	      for (i = curLevel->h - 1; i > y; i--, ptr -= curLevel->w)
		memcpy(ptr, ptr - curLevel->w, curLevel->w);
	    }

	    // Remove the mirror row if it exists and isn't the middle row
	    if (mirrorVert) {
	      y = (curLevel->h-1) - y;
	      if (y < (curLevel->h>>1)) {
		ptr = &maze[0];
		for (i = 0; i < y; i++, ptr += curLevel->w)
		  memcpy(ptr, ptr + curLevel->w, curLevel->w);
	      } else {
		ptr = &maze[curLevel->w * (curLevel->h - 1)];
		for (i = curLevel->h - 1; i > y; i--, ptr -= curLevel->w)
		  memcpy(ptr, ptr - curLevel->w, curLevel->w);
	      }
	      memset(&maze[y*curLevel->w], ' ', curLevel->w);
	    }

	    y = cursorY>>16;
	    memset(&maze[y*curLevel->w], ' ', curLevel->w);

	    newMaze = TRUE;
	    break;

	  case OPTION_REMOVE:
	    // Remove the current row
	    y = cursorY>>16;
	    ptr = &maze[y*curLevel->w];
	    if (!mirrorVert || (y < (curLevel->h>>1))) {
	      for (i = y; i > 0; i--, ptr -= curLevel->w)
		memcpy(ptr, ptr - curLevel->w, curLevel->w);
	    } else {
	      for (i = y; i < (curLevel->h-1); i++, ptr += curLevel->w)
		memcpy(ptr, ptr + curLevel->w, curLevel->w);
	    }
	    for (i = 0; i < curLevel->w; i++, ptr++) *ptr = ' ';

	    // Remove the mirror row if it exists and isn't the middle row
	    if (mirrorVert) {
	      y = (curLevel->h-1) - y;
	      ptr = &maze[y*curLevel->w];
	      if ((y < (curLevel->h>>1)) ||
		  ((y == (curLevel->h>>1)) && (curLevel->h & 0x1))) {
		for (i = y; i > 0; i--, ptr -= curLevel->w)
		  memcpy(ptr, ptr - curLevel->w, curLevel->w);
	      } else {
		for (i = y; i < (curLevel->h-1); i++, ptr += curLevel->w)
		  memcpy(ptr, ptr + curLevel->w, curLevel->w);
	      }
	      for (i = 0; i < curLevel->w; i++, ptr++) *ptr = ' ';
	    }

	    newMaze = TRUE;
	    break;
	}
      }

      if ((joyEdge & JOY_B) || (joyEdge & JOY_START)) {
	if (optionChoose) optionChoose = FALSE;
	else              optionMenu   = FALSE;
      }

      if (!optionChoose && (joyEdge & (CTRLR_DPAD_UP  ))) optionIndex--;
      if (!optionChoose && (joyEdge & (CTRLR_DPAD_DOWN))) optionIndex++;

      if (optionIndex < 0           ) optionIndex += MAX_OPTIONS;
      if (optionIndex >= MAX_OPTIONS) optionIndex -= MAX_OPTIONS;

    } else {

      if (joyEdge & (JOY_START)) optionMenu = TRUE;

      x = cursorX>>16;
      y = cursorY>>16;
      mazeChar = &maze[x + y * curLevel->w];

      for (i = 0; i < 4; i++) {
	switch (i) {
          case 0:
	    x = cursorX>>16;
	    y = cursorY>>16;
	    break;
          case 1:
	    if (!mirrorVert) continue;
	    x = cursorX>>16;
	    y = curLevel->h - (cursorY>>16) - 1;
	    break;
          case 2:
	    if (!mirrorHoriz) continue;
	    x = curLevel->w - (cursorX>>16) - 1;
	    y = cursorY>>16;
	    break;
          default:
	    if (!mirrorVert || !mirrorHoriz) continue;
	    x = curLevel->w - (cursorX>>16) - 1;
	    y = curLevel->h - (cursorY>>16) - 1;
	    break;
	}

	//
	// The A button draws/erases walls
	//
	if (((joyEdge & JOY_A) && !drag) ||
	    (((drag == DRAG_WALL_ON) || (drag == DRAG_WALL_OFF)) &&
	     (((prevX>>16) != (cursorX>>16)) || ((prevY>>16) != (cursorY>>16)) || i))) {
	  if (i == 0) {
	    if (!drag) {
	      if ((*mazeChar >= TILE_WALL_MIN) && (*mazeChar <= TILE_WALL_MAX))
		drag = DRAG_WALL_OFF;
	      else {
		switch (*mazeChar) {
	          case TILE_WALL:  drag = DRAG_WALL_OFF;  break;
	          case TILE_EMPTY:
	          case TILE_SLOW:  drag = DRAG_WALL_ON;   break;
	          default:         if (overwrite) drag = DRAG_WALL_ON;  break;
		}
	      }
	    }

	    if (drag == DRAG_WALL_ON) {
	      switch (*mazeChar) {
	        case TILE_SLOW:
	        case TILE_EMPTY:      *mazeChar = TILE_WALL;       break;
	        default:              if (overwrite) *mazeChar = TILE_WALL; break;
	      }
	    } else {
	      if ((*mazeChar >= TILE_WALL_MIN) && (*mazeChar <= TILE_WALL_MAX))
		*mazeChar = TILE_EMPTY;
	      switch (*mazeChar) {
	        case TILE_WALL:       *mazeChar = TILE_EMPTY;  break;
	        default:              if (overwrite) *mazeChar = TILE_EMPTY; break;
	      }
	    }
	  } else {
            switch (*mazeChar) {
              case TILE_WALL:
              case TILE_EMPTY:
              case TILE_SLOW:
                maze[x + (y * curLevel->w)] = *mazeChar;
                break;
            }
	  }

	  if (showChars) {
	    updateTileChar(curLevel, x, y, MazeBuf);
	    if (x > 0)               updateTileChar(curLevel, x-1, y, MazeBuf);
	    if (x < (curLevel->w-1)) updateTileChar(curLevel, x+1, y, MazeBuf);
	    if (y > 0)               updateTileChar(curLevel, x, y-1, MazeBuf);
	    if (y < (curLevel->h-1)) updateTileChar(curLevel, x, y+1, MazeBuf);
	  } else {
	    updateTile(curLevel, x, y, MazeBuf, 
		       &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
		       wallScale, pillScale, fruitScale);
	    if (x > 0)
	      updateTile(curLevel, x-1, y, MazeBuf, 
			 &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
			 wallScale, pillScale, fruitScale);
	    if (x < (curLevel->w-1))
	      updateTile(curLevel, x+1, y, MazeBuf, 
			 &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
			 wallScale, pillScale, fruitScale);
	    if (y > 0)
	      updateTile(curLevel, x, y-1, MazeBuf, 
			 &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
			 wallScale, pillScale, fruitScale);
	    if (y < (curLevel->h-1))
	      updateTile(curLevel, x, y+1, MazeBuf, 
			 &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
			 wallScale, pillScale, fruitScale);
	  }
	}

	//
	// Let the user choose the specific wall
	//
	/*
	if ((drag == DRAG_WALL_ON) || (drag == DRAG_WALL_OFF)) {
	  if (joyEdge & (JOY_C_UP | JOY_C_DOWN | JOY_C_LEFT | JOY_C_RIGHT))
	    if ((*mazeChar == TILE_WALL) || (*mazeChar == TILE_EMPTY) ||
		(overwrite && ((*mazeChar < TILE_WALL_MIN) || (*mazeChar > TILE_WALL_MAX))))
	      *mazeChar = 'A';

	  if (joyEdge & JOY_C_UP   ) *mazeChar += ((*mazeChar - 'A') & (1<<2)) ? (1<<2) : -(1<<2);
	  if (joyEdge & JOY_C_DOWN ) *mazeChar += ((*mazeChar - 'A') & (1<<0)) ? (1<<0) : -(1<<0);
	  if (joyEdge & JOY_C_LEFT ) *mazeChar += ((*mazeChar - 'A') & (1<<3)) ? (1<<3) : -(1<<3);
	  if (joyEdge & JOY_C_RIGHT) *mazeChar += ((*mazeChar - 'A') & (1<<1)) ? (1<<1) : -(1<<1);

	  if (showChars)
	    updateTileChar(curLevel, x, y, MazeBuf);
	  else
	    updateTile(curLevel, x, y, MazeBuf, 
		       &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
		       wallScale, pillScale, fruitScale);
	}
	*/

	//
	// The B button erases everything
	//
	if (((joyEdge & JOY_B) && !drag) ||
	    ((drag == DRAG_ERASE) &&
	     (((prevX>>16) != (cursorX>>16)) || ((prevY>>16) != (cursorY>>16)) || i))) {
	  if (i == 0) {
	    drag = DRAG_ERASE;
	    *mazeChar = TILE_EMPTY;
	  } else {
	    maze[x + (y * curLevel->w)] = *mazeChar;
	  }

	  if (showChars) {
	    updateTileChar(curLevel, x, y, MazeBuf);
	    if (x > 0)               updateTileChar(curLevel, x-1, y, MazeBuf);
	    if (x < (curLevel->w-1)) updateTileChar(curLevel, x+1, y, MazeBuf);
	    if (y > 0)               updateTileChar(curLevel, x, y-1, MazeBuf);
	    if (y < (curLevel->h-1)) updateTileChar(curLevel, x, y+1, MazeBuf);
	  } else {
	    updateTile(curLevel, x, y, MazeBuf, 
		       &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
		       wallScale, pillScale, fruitScale);
	    if (x > 0)
	      updateTile(curLevel, x-1, y, MazeBuf, 
			 &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
			 wallScale, pillScale, fruitScale);
	    if (x < (curLevel->w-1))
	      updateTile(curLevel, x+1, y, MazeBuf, 
			 &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
			 wallScale, pillScale, fruitScale);
	    if (y > 0)
	      updateTile(curLevel, x, y-1, MazeBuf, 
			 &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
			 wallScale, pillScale, fruitScale);
	    if (y < (curLevel->h-1))
	      updateTile(curLevel, x, y+1, MazeBuf, 
			 &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
			 wallScale, pillScale, fruitScale);
	  }
	}
	
	//
	// The C-DOWN button draws/erases pills
	//
	if (((joyEdge & JOY_C_DOWN) && !drag) ||
	    (((drag == DRAG_PILL_ON) || (drag == DRAG_PILL_OFF)) &&
	     (((prevX>>16) != (cursorX>>16)) || ((prevY>>16) != (cursorY>>16)) || i))) {
	  if (i == 0) {
	    if (!drag) {
	      switch (*mazeChar) {
	        case TILE_SLOW:
	        case TILE_EMPTY:     drag = DRAG_PILL_ON;   break;
	        case TILE_SLOW_PILL:
	        case TILE_PILL:      drag = DRAG_PILL_OFF;  break;
	        default:             if (overwrite) drag = DRAG_PILL_ON;  break;
	      }
	    }
	    
	    if (drag == DRAG_PILL_ON) {
	      switch (*mazeChar) {
	        case TILE_SLOW_POWERPILL:
	        case TILE_SLOW_FRUIT: if (overwrite) *mazeChar = TILE_SLOW_PILL;  break;
	        case TILE_SLOW:       *mazeChar = TILE_SLOW_PILL;  break;
	        case TILE_EMPTY:      *mazeChar = TILE_PILL;       break;
	        default:              if (overwrite) *mazeChar = TILE_PILL;       break;
	      }
	    } else if (drag == DRAG_PILL_OFF) {
	      switch (*mazeChar) {
	        case TILE_PILL:       *mazeChar = TILE_EMPTY;      break;
	        case TILE_SLOW_POWERPILL:
	        case TILE_SLOW_FRUIT: if (overwrite) *mazeChar = TILE_SLOW;  break;
	        case TILE_SLOW_PILL:  *mazeChar = TILE_SLOW;       break;
	        default:              if (overwrite) *mazeChar = TILE_EMPTY; break;
	      }
	    }
	  } else {
            switch (*mazeChar) {
              case TILE_EMPTY:
              case TILE_PILL:
              case TILE_SLOW_PILL:
                maze[x + (y * curLevel->w)] = *mazeChar;
                break;
            }
	  }
	  
	  if (showChars)
	    updateTileChar(curLevel, x, y, MazeBuf);
	  else
	    updateTile(curLevel, x, y, MazeBuf, 
		       &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
		       wallScale, pillScale, fruitScale);
	}
	
	//
	// The C-RIGHT button draws/erases fruit
	//
	if (((joyEdge & JOY_C_RIGHT) && !drag) ||
	    (((drag == DRAG_FRUIT_ON) || (drag == DRAG_FRUIT_OFF)) &&
	     (((prevX>>16) != (cursorX>>16)) || ((prevY>>16) != (cursorY>>16)) || i))) {
	  if (i == 0) {
	    if (!drag) {
	      switch (*mazeChar) {
	        case TILE_SLOW:
	        case TILE_EMPTY:     drag = DRAG_FRUIT_ON;   break;
	        case TILE_SLOW_FRUIT:
	        case TILE_FRUIT:     drag = DRAG_FRUIT_OFF;  break;
	        default:             if (overwrite) drag = DRAG_FRUIT_ON;  break;
	      }
	    }

	    if (drag == DRAG_FRUIT_ON) {
	      switch (*mazeChar) {
	        case TILE_SLOW_POWERPILL:
	        case TILE_SLOW_PILL:  if (overwrite) *mazeChar = TILE_SLOW_FRUIT;  break;
	        case TILE_SLOW:       *mazeChar = TILE_SLOW_FRUIT;  break;
	        case TILE_EMPTY:      *mazeChar = TILE_FRUIT;       break;
	        default:              if (overwrite) *mazeChar = TILE_FRUIT;       break;
	      }
	    } else if (drag == DRAG_FRUIT_OFF) {
	      switch (*mazeChar) {
	        case TILE_FRUIT:      *mazeChar = TILE_EMPTY;      break;
	        case TILE_SLOW_POWERPILL:
	        case TILE_SLOW_PILL:  if (overwrite) *mazeChar = TILE_SLOW;  break;
	        case TILE_SLOW_FRUIT: *mazeChar = TILE_SLOW;       break;
	        default:              if (overwrite) *mazeChar = TILE_EMPTY; break;
	      }
	    }

	  } else {
            switch (*mazeChar) {
              case TILE_EMPTY:
              case TILE_FRUIT:
              case TILE_SLOW_FRUIT:
                maze[x + (y * curLevel->w)] = *mazeChar;
                break;
            }
	  }

	  if (showChars)
	    updateTileChar(curLevel, x, y, MazeBuf);
	  else
	    updateTile(curLevel, x, y, MazeBuf, 
		       &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
		       wallScale, pillScale, fruitScale);
	}

	//
	// The C-UP button draws/erases power pills
	//
	if (((joyEdge & JOY_C_UP) && !drag) ||
	    (((drag == DRAG_POWERPILL_ON) || (drag == DRAG_POWERPILL_OFF)) &&
	     (((prevX>>16) != (cursorX>>16)) || ((prevY>>16) != (cursorY>>16)) || i))) {
	  if (i == 0) {
	    if (!drag) {
	      switch (*mazeChar) {
	        case TILE_SLOW:
	        case TILE_EMPTY:     drag = DRAG_POWERPILL_ON;   break;
	        case TILE_SLOW_POWERPILL:
	        case TILE_POWERPILL: drag = DRAG_POWERPILL_OFF;  break;
	        default:             if (overwrite) drag = DRAG_POWERPILL_ON;  break;
	      }
	    }

	    if (drag == DRAG_POWERPILL_ON) {
	      switch (*mazeChar) {
	        case TILE_SLOW_FRUIT:
	        case TILE_SLOW_PILL:  if (overwrite) *mazeChar = TILE_SLOW_POWERPILL;  break;
	        case TILE_SLOW:       *mazeChar = TILE_SLOW_POWERPILL;  break;
	        case TILE_EMPTY:      *mazeChar = TILE_POWERPILL;       break;
	        default:              if (overwrite) *mazeChar = TILE_POWERPILL;       break;
	      }
	    } else if (drag == DRAG_POWERPILL_OFF) {
	      switch (*mazeChar) {
	        case TILE_POWERPILL:      *mazeChar = TILE_EMPTY;      break;
	        case TILE_SLOW_FRUIT:
	        case TILE_SLOW_PILL:  if (overwrite) *mazeChar = TILE_SLOW;  break;
	        case TILE_SLOW_POWERPILL: *mazeChar = TILE_SLOW;       break;
	        default:              if (overwrite) *mazeChar = TILE_EMPTY; break;
	      }
	    }

	  } else {
            switch (*mazeChar) {
              case TILE_EMPTY:
              case TILE_POWERPILL:
              case TILE_SLOW_POWERPILL:
                maze[x + (y * curLevel->w)] = *mazeChar;
                break;
            }
	  }

	  if (showChars)
	    updateTileChar(curLevel, x, y, MazeBuf);
	  else
	    updateTile(curLevel, x, y, MazeBuf, 
		       &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
		       wallScale, pillScale, fruitScale);
	}

	//
	// The C-LEFT button toggles slow tile
	//
	if (((joyEdge & JOY_C_LEFT) && !drag) ||
	    (((drag == DRAG_SLOW_ON) || (drag == DRAG_SLOW_OFF)) &&
	     (((prevX>>16) != (cursorX>>16)) || ((prevY>>16) != (cursorY>>16)) || i))) {
	  if (i == 0) {
	    if (!drag) {
	      switch (*mazeChar) {
	        case TILE_PILL:
	        case TILE_POWERPILL:
	        case TILE_FRUIT:
	        case TILE_EMPTY:     drag = DRAG_SLOW_ON;   break;
	        case TILE_SLOW_PILL:
	        case TILE_SLOW_POWERPILL:
	        case TILE_SLOW_FRUIT:
	        case TILE_SLOW:      drag = DRAG_SLOW_OFF;  break;
	      }
	    }

	    if (drag == DRAG_SLOW_ON) {
	      switch (*mazeChar) {
	        case TILE_POWERPILL:  *mazeChar = TILE_SLOW_POWERPILL;  break;
	        case TILE_PILL:       *mazeChar = TILE_SLOW_PILL;       break;
	        case TILE_FRUIT:      *mazeChar = TILE_SLOW_FRUIT;      break;
	        case TILE_EMPTY:      *mazeChar = TILE_SLOW;            break;
	      }
	    } else if (drag == DRAG_SLOW_OFF) {
	      switch (*mazeChar) {
	        case TILE_SLOW_POWERPILL:  *mazeChar = TILE_POWERPILL;  break;
	        case TILE_SLOW_PILL:       *mazeChar = TILE_PILL;       break;
	        case TILE_SLOW_FRUIT:      *mazeChar = TILE_FRUIT;      break;
	        case TILE_SLOW:            *mazeChar = TILE_EMPTY;      break;
	      }
	    }

	  } else {
            switch (*mazeChar) {
              case TILE_EMPTY:
              case TILE_SLOW:
              case TILE_PILL:
              case TILE_SLOW_PILL:
              case TILE_FRUIT:
              case TILE_SLOW_FRUIT:
              case TILE_POWERPILL:
              case TILE_SLOW_POWERPILL:
                maze[x + (y * curLevel->w)] = *mazeChar;
                break;
            }
	  }

	  if (showChars)
	    updateTileChar(curLevel, x, y, MazeBuf);
	  else
	    updateTile(curLevel, x, y, MazeBuf, 
		       &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
		       wallScale, pillScale, fruitScale);
	}

      }

      // Restore X & Y
      x = cursorX>>16;
      y = cursorY>>16;

      //
      // L button adds/deletes Chomper start position
      //
      if ((joyEdge & JOY_L) && !drag && !(joystick & JOY_NUON)) {
	// Redraw maze if we nuked a wall
	if (overwrite && ((*mazeChar == TILE_WALL) ||
			  ((*mazeChar >= TILE_WALL_MIN) && (*mazeChar <= TILE_WALL_MAX))))
	  newMaze = TRUE;

        if (showChars) newMaze = TRUE;
	if (*mazeChar == 'X') {
	  *mazeChar = ' ';
	} else if (overwrite || (*mazeChar == ' ')) {
	  for (ptr = maze; *ptr && (*ptr != 'X'); ptr++);
	  if (*ptr == 'X') *ptr = ' ';
	  *mazeChar = 'X';
	}

	if (showChars)
	  updateTileChar(curLevel, x, y, MazeBuf);
	else
	  updateTile(curLevel, x, y, MazeBuf, 
		     &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
		     wallScale, pillScale, fruitScale);
      }

      //
      // R button adds/deletes enemy start position
      //
      if ((joyEdge & JOY_R) && !drag && !(joystick & JOY_NUON)) {
	if ((*mazeChar >= TILE_ENEMY_MIN) && (*mazeChar <= TILE_ENEMY_MAX)) {
	  // Deleting an enemy
	  *mazeChar = ' ';

	} else {
	  // Adding an enemy

	  // Find out how many enemies are already in the maze
	  for (ptr = maze, i = 0; *ptr; ptr++)
	    if ((*ptr >= TILE_ENEMY_MIN) && (*ptr <= TILE_ENEMY_MAX))
	      i |= 1<<(*ptr - TILE_ENEMY_MIN);

	  // Only add enemy if not at max
	  if (i != ((1<<MAX_GHOSTS)-1)) {
	    if (overwrite || (*mazeChar == ' ')) {
	      // Redraw maze if we nuked a wall
	      if ((*mazeChar == TILE_WALL) ||
		  ((*mazeChar >= TILE_WALL_MIN) && (*mazeChar <= TILE_WALL_MAX)))
		newMaze = TRUE;
	      for (*mazeChar = TILE_ENEMY_MIN; i & 0x1; i >>= 1, (*mazeChar)++);
	    }
	  }
	}

	// Redraw the tile
	if (showChars)
	  updateTileChar(curLevel, x, y, MazeBuf);
	else
	  updateTile(curLevel, x, y, MazeBuf, 
		     &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex], 
		     wallScale, pillScale, fruitScale);
      }

      if (joyEdge & JOY_NUON) {
	showChars = 1 - showChars;
	newMaze = TRUE;
      }

      if ( (joystick & JOY_NUON) && (emulator || DSI_DEBUG) &&
           ( ((joyEdge & JOY_L) && (joystick & JOY_R)) ||
             ((joyEdge & JOY_R) && (joystick & JOY_L)) )) {
        mazeChar = maze;
        printf("You can cut-and-paste this text directly into one of the structures");
        printf("in 'levels.h'\n");
        printf("  {\n");
        printf("    \"");
        for (x = 0; x < curLevel->w; x++)
          printf("-");
        printf("\",\n");
        for (y = 0; y < curLevel->h; y++) {
          printf("    \"");
          for (x = 0; x < curLevel->w; x++, mazeChar++)
            printf("%c", *mazeChar);
          printf("\"\n");
        }
        printf("  },\n");
        fflush(stdout);
      }
    }

    SPRSetSpriteXY(cursor0, SPRITE_SCREEN_X((cursorX & 0xFFFF0000) | 0x8000),
		   SPRITE_SCREEN_Y((cursorY & 0xFFFF0000) | 0x8000));

    //
    // Draw
    //

    // Draw chomperman if the start position exists
    for (ptr = maze, x = 0; *ptr && (*ptr != 'X'); ptr++, x++);
    if (!showChars) {
      if (*ptr == 'X') {
	y = x / curLevel->w;
	x = x % curLevel->w;
	chomper.tileX = (x << 16) | 0x8000;
	chomper.tileY = (y << 16) | 0x8000;
	SPRAddSprite(chomper.chomperman,
		     SPRITE_SCREEN_X(chomper.tileX),
		     SPRITE_SCREEN_Y(chomper.tileY),
		     0,
		     CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		     CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		     0x40000002, kBlack, 1);
      }
    }

    // Draw enemies if they exist
    if (!showChars) {
      for (ptr = maze, x = y = 0; *ptr; ptr++) {
	if ((*ptr >= TILE_ENEMY_MIN) && (*ptr <= TILE_ENEMY_MAX)) {
	  ghosts[*ptr - TILE_ENEMY_MIN].startX = x;
	  ghosts[*ptr - TILE_ENEMY_MIN].startY = y;
	  ghosts[*ptr - TILE_ENEMY_MIN].tileX = (x << 16) | 0x8000;
	  ghosts[*ptr - TILE_ENEMY_MIN].tileY = (y << 16) | 0x8000;
	  SPRAddSprite(ghosts[*ptr - TILE_ENEMY_MIN].spr,
		       SPRITE_SCREEN_X(ghosts[*ptr - TILE_ENEMY_MIN].tileX),
		       SPRITE_SCREEN_Y(ghosts[*ptr - TILE_ENEMY_MIN].tileY),
		       0,
		       aGhostImg[ghostIndex].scale * GHOSTSCALE,
		       aGhostImg[ghostIndex].scale * GHOSTSCALE,
		       0x40000002, aGhostImg[ghostIndex].transColor, 1);
	}
	x++;
	if (x >= curLevel->w) { x = 0; y++; }
      }
    }

    // Show other cursors if mirrored
    if (mirrorVert) {
      x = (cursorX & 0xFFFF0000) | 0x8000;
      y = (((curLevel->h-1) - (cursorY>>16)) << 16) | 0x8000;
      if ((y>>16) != (cursorY>>16))
	SPRAddSprite(cursor1, SPRITE_SCREEN_X(x), SPRITE_SCREEN_Y(y),
		     0, 0x10000, 0x10000, kSpriteInterpolatedTrans | (0x04<<26), 0, 2);
    }
    if (mirrorHoriz) {
      x = (((curLevel->w-1) - (cursorX>>16)) << 16) | 0x8000;
      y = (cursorY & 0xFFFF0000) | 0x8000;
      if ((x>>16) != (cursorX>>16))
	SPRAddSprite(cursor2, SPRITE_SCREEN_X(x), SPRITE_SCREEN_Y(y),
		     0, 0x10000, 0x10000, kSpriteInterpolatedTrans | (0x04<<26), 0, 2);
    }
    if (mirrorVert && mirrorHoriz) {
      x = (((curLevel->w-1) - (cursorX>>16)) << 16) | 0x8000;
      y = (((curLevel->h-1) - (cursorY>>16)) << 16) | 0x8000;
      if (((x>>16) != (cursorX>>16)) && ((y>>16) != (cursorY>>16)))
	SPRAddSprite(cursor3, SPRITE_SCREEN_X(x), SPRITE_SCREEN_Y(y),
		     0, 0x10000, 0x10000, kSpriteInterpolatedTrans | (0x04<<26), 0, 2);
    }

    // Draw a new maze?
    if (newMaze) {
      newMaze = FALSE;
      
      // Make sure all sprites are properly installed and initialized
      checkAndLoadTGA(&aWallImg [wallIndex ], 5, TRUE);
      checkAndLoadTGA(&aPillImg [pillIndex ], 1, TRUE);
      checkAndLoadTGA(&aFruitImg[fruitIndex], 1, TRUE);
      checkAndLoadTGA(&aGhostImg[ghostIndex], 7, TRUE);

      wallScale  = (TILESIZE * 0x10000 * aWallImg [wallIndex ].scale) / SPRITESIZE;
      pillScale  = (TILESIZE * 0x10000 * aPillImg [pillIndex ].scale) / SPRITESIZE;
      fruitScale = (TILESIZE * 0x10000 * aFruitImg[fruitIndex].scale) / SPRITESIZE;

      // Get new offset in case dimensions changed
      curLevel->offsetX = (MAZE_MAX_W - curLevel->w) * (TILESIZE>>1);
      curLevel->offsetY = (MAZE_MAX_H - curLevel->h) * (TILESIZE>>1);

      // Draw the maze
      myFillRect(MazeBuf, DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);
      if (showChars)
	CreateMaze(curLevel, 0, 0, 0, MazeBuf);
      else
	CreateMaze(curLevel, &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex],
		   MazeBuf);

      /*
      // Make sure enemy graphic set is loaded
      checkAndLoadTGA(&aGhostImg[ghostIndex], 7, TRUE);

      // Need to delete and recreate ghosts since sprite could change
      for(i = 0; i < MAX_GHOSTS; i++) {
	if (ghosts[i].spr) {
	  SPRRemoveSprite(ghosts[i].spr);
	  SPRDeleteSprite(ghosts[i].spr);
	}
	ghosts[i].spr = SPRCreateSprite(&aGhostImg[ghostIndex].spr,
					i * aGhostImg[ghostIndex].w, 0,
					aGhostImg[ghostIndex].w, aGhostImg[ghostIndex].w);
      }

      for(i = 0; i < MAX_GHOSTS; i++) {
	if (ghosts[i].startX >= 0) {
	  ghosts[i].tileX  = (ghosts[i].startX << 16) | 0x8000;
	  ghosts[i].tileY  = (ghosts[i].startY << 16) | 0x8000;
	  SPRAddSprite(ghosts[i].spr,
		       SPRITE_SCREEN_X(ghosts[i].tileX),
		       SPRITE_SCREEN_Y(ghosts[i].tileY),
		       0,
		       aGhostImg[ghostIndex].scale * GHOSTSCALE,
		       aGhostImg[ghostIndex].scale * GHOSTSCALE,
		       0x40000002, aGhostImg[ghostIndex].transColor, 1);
	}
      }

      chomper.tileX = (chomper.startX << 16) | 0x8000;
      chomper.tileY = (chomper.startY << 16) | 0x8000;
      chomper.aphase = 1;
      SPRSetSpriteXY(chomper.chomperman,
		     SPRITE_SCREEN_X(chomper.tileX),
		     SPRITE_SCREEN_Y(chomper.tileY));
      SPRSetSpriteSource(chomper.chomperman,
			 (chomper.aphase * SPRITESIZE) / aPlayerImg[playerIndex].scale, 0,
			 aPlayerImg[playerIndex].w, aPlayerImg[playerIndex].w);
      */
    }

    myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);

    // Draw maze
    CopyMaze(FrameBuffer[drawBuffer], MazeBuf, SCREEN_WIDTH * SCREEN_HEIGHT * 4 / 256);

    SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH - 1,
		     SCREEN_HEIGHT - 1, kBlack);
    SPRDraw(0,0);
    SPRWait();

    // Remove chomperman if the start position exists
    if (!showChars) {
      for (ptr = maze; *ptr && (*ptr != 'X'); ptr++);
      if (*ptr == 'X') SPRRemoveSprite(chomper.chomperman);
    }

    // Remove enemies if they exist
    if (!showChars) {
      for (ptr = maze; *ptr; ptr++)
	if ((*ptr >= TILE_ENEMY_MIN) && (*ptr <= TILE_ENEMY_MAX))
	  SPRRemoveSprite(ghosts[*ptr - TILE_ENEMY_MIN].spr);
    }

    if (mirrorVert) SPRRemoveSprite(cursor1);
    if (mirrorHoriz) SPRRemoveSprite(cursor2);
    if (mirrorVert && mirrorHoriz) SPRRemoveSprite(cursor3);

    // Draw the level boundaries
    SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
               96, 16, 8, 8, FrameBuffer[drawBuffer], DMAFLAGS,
	       SPRITE_SCREEN_X(0x8000) - 2, SPRITE_SCREEN_Y(0x8000) - 2,
	       0x10000, 0x10000, 0,
	       0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<29),0);

    SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
               104, 16, 8, 8, FrameBuffer[drawBuffer], DMAFLAGS,
	       SPRITE_SCREEN_X(((curLevel->w-1)<<16) | 0x8000) + 2, SPRITE_SCREEN_Y(0x8000) - 2,
	       0x10000, 0x10000, 0,
	       0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<29),0);

    SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
               96, 24, 8, 8, FrameBuffer[drawBuffer], DMAFLAGS,
	       SPRITE_SCREEN_X(0x8000) - 2, SPRITE_SCREEN_Y(((curLevel->h-1)<<16) | 0x8000) + 2,
	       0x10000, 0x10000, 0,
	       0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<29),0);

    SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
               104, 24, 8, 8, FrameBuffer[drawBuffer], DMAFLAGS,
	       SPRITE_SCREEN_X(((curLevel->w-1)<<16) | 0x8000) + 2,
	       SPRITE_SCREEN_Y(((curLevel->h-1)<<16) | 0x8000) + 2,
	       0x10000, 0x10000, 0,
	       0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<29),0);

    // Draw option menu
    if (optionMenu) {
      x = 30;
      y = 15;
      checkAndLoadTGA(&aImage[kImgLevelEditOption], 1, TRUE);
      SPRBlitter(aImage[kImgLevelEditOption].spr.img, aImage[kImgLevelEditOption].spr.dmaflags,
                 0, 0, 128, 256,
		 FrameBuffer[drawBuffer], DMAFLAGS, (128>>1)+x, (256>>1)+y, 0x10000, 0x10000, 0,
		 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<30), 0);

      y = 16 + optionIndex*17;
      if (optionChoose)
	SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
                   0, 0, 128, 16,
		   FrameBuffer[drawBuffer], DMAFLAGS, (128>>1)+x, (16>>1)+y, 0x10000, 0x10000, 0,
		   0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<29),0);
      else
	SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
                   0, 0, 128, 16,
		   FrameBuffer[drawBuffer], DMAFLAGS, (128>>1)+x, (16>>1)+y, 0x10000, 0x10000, 0,
		   0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<30),0);

      y = 15 + 17*OPTION_MAZE_W;
      i = ((optionIndex == OPTION_MAZE_W) && optionChoose) ? newValue : curLevel->w;
      if (curLevel->w > 9)
	SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
                   16+8*(i/10), 16, 8, 16,
                   FrameBuffer[drawBuffer], DMAFLAGS, (8>>1)+x, (16>>1)+y, 0x10000, 0x10000, 0,
		   0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<30),0);
      SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
                 16+8*(i%10), 16, 8, 16,
                 FrameBuffer[drawBuffer], DMAFLAGS, (8>>1)+x+6, (16>>1)+y, 0x10000, 0x10000,0,
		 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<30),0);

      y = 15 + 17*OPTION_MAZE_H;
      i = ((optionIndex == OPTION_MAZE_H) && optionChoose) ? newValue : curLevel->h;
      if (i > 9)
	SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
                   16+8*(i/10), 16, 8, 16,
                   FrameBuffer[drawBuffer], DMAFLAGS, (8>>1)+x, (16>>1)+y, 0x10000, 0x10000, 0,
		   0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<30),0);
      SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
                 16+8*(i%10), 16, 8, 16,
                 FrameBuffer[drawBuffer], DMAFLAGS, (8>>1)+x+6, (16>>1)+y,0x10000, 0x10000,0,
		 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<30),0);

      if (mirrorHoriz) {
	y = 15 + 17*OPTION_MIRROR_HORIZ;
	SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
                   0, 16, 16, 16,
		   FrameBuffer[drawBuffer], DMAFLAGS, (16>>1)+x, (16>>1)+y, 0x10000, 0x10000, 0,
		   0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<30),0);
      }
      if (mirrorVert) {
	y = 15 + 17*OPTION_MIRROR_VERT;
	SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
                   0, 16, 16, 16,
		   FrameBuffer[drawBuffer], DMAFLAGS, (16>>1)+x, (16>>1)+y, 0x10000, 0x10000, 0,
		   0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<30),0);
      }
      if (overwrite) {
	y = 15 + 17*OPTION_OVERWRITE;
	SPRBlitter(aImage[kImgLevelEditOther].spr.img, aImage[kImgLevelEditOther].spr.dmaflags,
                   0, 16, 16, 16,
		   FrameBuffer[drawBuffer], DMAFLAGS, (16>>1)+x, (16>>1)+y, 0x10000, 0x10000, 0,
		   0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, kSpriteInterpolatedTrans | (1<<30),0);
      }
    }

    VideoSync();
  }


  // Remove Chomperman and enemies
  for(i = 0; i < MAX_GHOSTS; i++) {
    SPRDeleteSprite(ghosts[i].spr);
    ghosts[i].spr = NULL;
  }

  SPRRemoveSprite(chomper.chomperman);

  SPRRemoveSprite(cursor0);
  SPRDeleteSprite(cursor0);
  SPRDeleteSprite(cursor1);
  SPRDeleteSprite(cursor2);
  SPRDeleteSprite(cursor3);

  // Make changes permanent
  memcpy(curLevel->maze, maze, curLevel->w * curLevel->h);

  // Compute time it would take Chomper to traverse every non-wall tile
  curLevel->distance = 0;
  for (ptr = curLevel->maze; *ptr; ptr++)
    if ((*ptr != TILE_WALL) && ((*ptr < TILE_WALL_MIN) || (*ptr > TILE_WALL_MAX)))
      curLevel->distance += 0x10000;
  curLevel->distance /= SPEED_CHOMPER;

  // Make width calibration correct
  for (i = 0, ptr = curLevel->widthCalib; i < curLevel->w; i++, ptr++) *ptr = ' ';
  *ptr = 0;
  
  return(returnValue);
}

//==================================================================================================
//==================================================================================================

void levelDone()
{
  char tmpString[9];
  int i, x, y;
  int newPercent = 0;

  // Quiet things down
  if (!mute) PCMVoiceOff(voiceChase);

  // Advance to next level unless we're in debug mode
  if ((gameMode == MODE_ARCADE) || (gameMode == MODE_ARCADE_RAND)) curLevelIndex++;

  // Remove spinning score sprite
  if (saphase) {  SPRRemoveSprite(bonusSprite);  saphase = 0;  }

  // Remove all smoke
  resetSmoke();

  // Reset chomper
  chomper.angle = 0;

  // Reset Chomper orientation
  SPRSetSpriteRotation(chomper.chomperman, 0);
  SPRSetSpriteScale(chomper.chomperman,
		    CHOMPERSCALE * aPlayerImg[playerIndex].scale,
		    CHOMPERSCALE * aPlayerImg[playerIndex].scale);

  // SDC 2002-05-26:  Make enemies speed up every level
  if (gameMode != MODE_TOURNAMENT)
    for (i = 0; i < MAX_GHOSTS; i++)
      if (ghosts[i].speed < SPEED_ENEMY_MAX)
        ghosts[i].speed += SPEED_ENEMY_INC;

  // Check if done with last level
  if (((gameMode==MODE_ARCADE) || (gameMode==MODE_ARCADE_RAND)) && (curLevelIndex == numLevels)) {
    checkAndLoadTGA(&aImage[kImgGameComplete], 1, TRUE);
    SPRBlitter(aImage[kImgGameComplete].spr.img, aImage[kImgGameComplete].spr.dmaflags,
               0, 0, 256, 128, FrameBuffer[drawBuffer], DMAFLAGS, SCREEN_WIDTH/2, SCREEN_HEIGHT/2,
	       0x18000 * SCREEN_WIDTH / 720, 0x18000 * SCREEN_HEIGHT / 480, 0,
	       0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);	      
  } else {
    checkAndLoadTGA(&aImage[kImgLevelComplete], 1, TRUE);
    SPRBlitter(aImage[kImgLevelComplete].spr.img, aImage[kImgLevelComplete].spr.dmaflags,
               0, 0, 256, 128, FrameBuffer[drawBuffer], DMAFLAGS, 
	       SCREEN_WIDTH/2, (gameMode == MODE_TOURNAMENT) ? 50 : SCREEN_HEIGHT/2,
	       0x18000 * SCREEN_WIDTH / 720, 0x18000 * SCREEN_HEIGHT / 480, 0,
	       0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
  }

  // SDC 2002-06-07:  Display adjusted score in tournament mode based on time
  if (gameMode == MODE_TOURNAMENT) {
    if (percent >= curLevel->distance) {
      newPercent = 0;
    } else {
      newPercent = ((curLevel->distance - percent) << 16) / curLevel->distance;
      newPercent >>= 4;
    }
	    
    x = (SCREEN_WIDTH  / 2) - 100;
    y = (SCREEN_HEIGHT / 2) + 20;
    x = drawString("level score  \0", FrameBuffer[drawBuffer], x, y, 0x40000000);
    sprintf(tmpString, "%6d", score);
    drawString(tmpString, FrameBuffer[drawBuffer], x, y, 0x40000000);

    x = (SCREEN_WIDTH  / 2) - 100;
    y += 2 * (24 * SCORESCALE) / 0x10000;
    x = drawString("time percent \0", FrameBuffer[drawBuffer], x, y, 0x40000000);
    sprintf(tmpString, "%3d", (newPercent * 100) >> 12);
    x = drawString(tmpString, FrameBuffer[drawBuffer], x, y, 0x40000000);
    x = drawString(".\0", FrameBuffer[drawBuffer], x, y, 0x40000000);
    sprintf(tmpString, "%02d", ((newPercent * 10000) >> 12) % 100);
    x = drawString(tmpString, FrameBuffer[drawBuffer], x, y, 0x40000000);
    drawString("%\0", FrameBuffer[drawBuffer], x, y, 0x40000000);

    score *= newPercent;
    score >>= 12;

    x = (SCREEN_WIDTH  / 2) - 100;
    y += 2 * (24 * SCORESCALE) / 0x10000;
    drawString("final score\0", FrameBuffer[drawBuffer], x, y, 0x40000000);
    x += 13 * (24 * SCORESCALE) / 0x10000;
    sprintf(tmpString, "%6d", score);
    drawString(tmpString, FrameBuffer[drawBuffer], x, y, 0x40000000);
  }

  myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, 0, SCREEN_HEIGHT - 32, SCREEN_WIDTH, 32, kBlack);

  // Update final score
  drawScore(score, FrameBuffer[drawBuffer], SCOREX, SCOREY, 0x40000000,
	    (gameMode == MODE_TOURNAMENT) ? 0 : 1);

  VideoSync();
  WaitForJoystick(JOY_A);

  if (gameMode == MODE_TOURNAMENT) {
    if (score > aLevelScore[levelTournament].score) {
      getName(curName);
      strncpy(aLevelScore[levelTournament].name, curName, 8);
      aLevelScore[levelTournament].score   = score;
      aLevelScore[levelTournament].percent = percent;
      showVerfCode(levelTournament);
    } else {
      // Consolation
    }
  }

  while ((_Controller[1].buttons | _Controller[0].buttons) & (JOY_A | JOY_B));
}

//==================================================================================================
//==================================================================================================

/*
  Main loop
*/

int main()
{
  int i, next, pinc;
  VidDisplay display;
  VidChannel mainchannel;
  char *ptr;
  struct BiosInfo *bios;
  
  //  _MediaShutdownMPE();
  //  _MediaInitMPE();

  bios = _BiosGetInfo();
  if (!strncmp(bios->info_string, "Nuance", strlen("Nuance"))) emulator = TRUE;
  if (emulator) {
    printf("CHOMP v%s\n", versionString);
    printf("-----------");
    printf("\nEmulator detected");
    printf("\nTournament scores are invalid using the emulator");
    fflush(stdout);
  }

  mute = emulator;
  
  // Setup audio
  if (!mute) { initAudio = TRUE;  AUDIOInit(); }

  Pan.PCMPanLR = 0;
  Pan.PCMPanFB = 0;
  Pan.PCMPanUD = 0;
  
  // Get some local memory for DMA
  
  localScratch = _MemLocalScratch(&localSize);
  
  // Setup video
  
  for(i = 0; i < NUM_BUFFERS; i++)
    FrameBuffer[i]=_MemAlloc(SCREEN_WIDTH * SCREEN_HEIGHT * 4, 512, kMemSDRAM);
  MazeBuf = _MemAlloc(SCREEN_WIDTH * SCREEN_HEIGHT * 4, 512, kMemSDRAM);
  
  myFillRect(MazeBuf, DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);
  
  display.dispwidth   = -1;
  display.dispheight  = -1;
  display.bordcolor   = 0x20100000;
  display.progressive = 0;
  for(i = 0; i < 6; i++) display.reserved[i] = 0;
  
  mainchannel.dmaflags    = DMAFLAGS;
  mainchannel.base        = FrameBuffer[0];
  mainchannel.dest_xoff   = -1;
  mainchannel.dest_yoff   = -1;
  mainchannel.dest_width  = 720;
  mainchannel.dest_height = 480;
  mainchannel.src_xoff    = 0;
  mainchannel.src_yoff    = 0;
  mainchannel.src_width   = SCREEN_WIDTH;
  mainchannel.src_height  = SCREEN_HEIGHT;
  mainchannel.clut_select = 0;
  mainchannel.alpha       = 0;
  mainchannel.vfilter     = VID_VFILTER_2TAP;
  mainchannel.hfilter     = VID_HFILTER_4TAP;
  for(i = 0; i < 5; i++) mainchannel.reserved[i] = 0;
  
  InitTerminalX(0,(int)FrameBuffer[0],SCREEN_WIDTH, SCREEN_HEIGHT, DMAFLAGS, kBlack);
  _VidConfig ( &display, &mainchannel, 0, 0L);
  
  initRandomArray();

  srand(_VidSync(-1));
  
  // Use MPEs 0,1 and 2 for rendering, 16 pixel slices

  if (emulator)
    SPRInit(2, 2, 240, 0);
  else
    SPRInit(0, 2, 16, 0);

  // Set to the second framebuffer
  SPRSetDestScreen(FrameBuffer[1], DMAFLAGS,  0,  0,  SCREEN_WIDTH-1,  SCREEN_HEIGHT-1, kBlack);
  
  myFillRect(FrameBuffer[0], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);
  myFillRect(FrameBuffer[1], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);

  // SDC 2002-02-10:  Random sprites
  // Find out how many of each type we have
  playerIndex = 0;
  for (playerMax = 0; aPlayerImg[playerMax].tgaAddr; playerMax++ ) {
    aPlayerImg[wallMax].init = FALSE;
    checkAndLoadTGA(&aPlayerImg[playerMax], 4, TRUE);
  }
  for (wallMax  = 0; aWallImg [wallMax ].tgaAddr; wallMax++ ) aWallImg[wallMax].init   = FALSE;
  for (pillMax  = 0; aPillImg [pillMax ].tgaAddr; pillMax++ ) aPillImg[pillMax].init   = FALSE;
  for (fruitMax = 0; aFruitImg[fruitMax].tgaAddr; fruitMax++) aFruitImg[fruitMax].init = FALSE;
  for (ghostMax = 0; aGhostImg[ghostMax].tgaAddr; ghostMax++) aGhostImg[ghostMax].init = FALSE;

  // Install the TGA images

  checkAndLoadTGA(&aImage[kImgMisc], 4, TRUE);
  checkAndLoadTGA(&aImage[kImgFont], 8, TRUE);

  // Create sprites
  
  chomper.chomperman = SPRCreateSprite(&aPlayerImg[playerIndex].spr, aPlayerImg[playerIndex].w, 0,
				       aPlayerImg[playerIndex].w, aPlayerImg[playerIndex].w);
  for(i = 0; i < MAX_GHOSTS; i++) ghosts[i].spr = NULL;

  // Initialize maze info
  initMazes();

  for (i = 0; i < MAX_SMOKE; i++)
    aSmoke[i].spr = SPRCreateSprite(&aImage[kImgMisc].spr,
				    SMOKE_OFFSET_X * SPRITESIZE, SMOKE_OFFSET_Y * SPRITESIZE,
				    aImage[kImgMisc].w, aImage[kImgMisc].w);
  smokeHead = smokeTail = 0;
  resetSmoke();

  // main game loop
  
  for(;;) {
     
    if (gameMode == MODE_START) {
      // Start screen

      myFillRect(FrameBuffer[0], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);
      _VidChangeBase(VID_CHANNEL_MAIN,DMAFLAGS,FrameBuffer[0]);
      
      checkAndLoadTGA(&aImage[kImgStartScreen], 1, TRUE);
      SPRBlitter(aImage[kImgStartScreen].spr.img, aImage[kImgStartScreen].spr.dmaflags,
                 0, 0, 600, 400, FrameBuffer[0], DMAFLAGS,
		 SCREEN_WIDTH/2, SCREEN_HEIGHT/2,
		 0x10000 * SCREEN_WIDTH / 720, 0x10000 * SCREEN_HEIGHT / 480,
		 0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
      
      WaitForJoystick(JOY_START);
      while (ButtonStart(_Controller[1]));
    }

    gameMode = getGameMode();
    randomizeMazes();

    //    debug = ((_Controller[1].buttons | _Controller[0].buttons) & JOY_B) ? 1 : 0;

    score = 0;
    lives = 2;
    extra = 10000;  // next extra life at 10000 points

    if ((gameMode == MODE_ARCADE) || (gameMode == MODE_ARCADE_RAND)) {
      myPlayWAV((void *) GameBeginWav, -1, &Pan, 0x40000000, 0x40000000);
      for(i = 0; i < MAX_GHOSTS; i++) ghosts[i].speed = SPEED_ENEMY;
      curLevelIndex = 0;
    } else {
      for(i = 0; i < MAX_GHOSTS; i++)
	ghosts[i].speed = SPEED_ENEMY + (((SPEED_ENEMY_MAX - SPEED_ENEMY) * 0xC000) >> 16);
    }

    // Level loop
      
    while (1) {
      if ((gameMode == MODE_SELECT) || (gameMode == MODE_START)) break;

      if (((gameMode==MODE_ARCADE) || (gameMode==MODE_ARCADE_RAND)) && (curLevelIndex >= numLevels))
	break;

      collected = 0;
      delay     = 0;
      gameover  = 0;

      if (gameMode == MODE_LEVEL_SELECT) {
	curLevelIndex = selectLevel(levels, curLevelIndex, numLevels);
	if (curLevelIndex == -1) break;
	score = 0;
	lives = 2;
	extra = 10000;  // next extra life at 10000 points
      } else if (gameMode == MODE_TOURNAMENT) {
	levelTournament = selectLevel(aLevelTournament, levelTournament, numLevelsTournament);
	if (levelTournament == -1) break;
	score = 0;
	lives = 2;
	extra = 10000;  // next extra life at 10000 points
      } else if (gameMode == MODE_LEVEL_EDITOR) {
	curLevelIndex = editLevel(0);
	if (curLevelIndex == -1) break;
	score = 0;
	lives = 2;
	extra = 10000;  // next extra life at 10000 points
      }

      // Get pointer to current level
      switch (gameMode) {
        case MODE_ARCADE:
        case MODE_LEVEL_SELECT:
	  curLevel = &levels[curLevelIndex];
	  break;

        case MODE_ARCADE_RAND:
	  curLevel = &levels[aRandMazes[curLevelIndex]];
	  break;

        case MODE_TOURNAMENT:
	  curLevel = &aLevelTournament[levelTournament];
	  break;

        case MODE_LEVEL_EDITOR:
	  curLevel = &aLevelEdited[curLevelIndex];
	  break;
      }
      memcpy(maze, curLevel->maze, curLevel->w * curLevel->h);

      if ((gameMode == MODE_TOURNAMENT) || (gameMode == MODE_LEVEL_EDITOR)) {
	percentDelay = curLevel->distance;
	percent = 0;
        if (!mute) {
          myPlayWAV((void *) GetReadyWav, -1, &Pan, 0x40000000, 0x40000000);
          while (PCMGetUsedVoices());
        }
      }

      // Clear level before drawing
      myFillRect(MazeBuf, DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);

      if ((gameMode == MODE_ARCADE) || (gameMode == MODE_ARCADE_RAND)) {
	// SDC 2002-02-10:  Choose sprites at random for each level
	wallIndex  = rand() % wallMax;
	pillIndex  = rand() % pillMax;
	fruitIndex = rand() % fruitMax;
	ghostIndex = rand() % ghostMax;

	// Make sure all sprites are properly installed and initialized
        EMU_PRINT("\nLoading level images: wall");
	checkAndLoadTGA(&aWallImg [wallIndex ], 5, FALSE);
        EMU_PRINT(" pill");
	checkAndLoadTGA(&aPillImg [pillIndex ], 1, FALSE);
        EMU_PRINT(" fruit");
	checkAndLoadTGA(&aFruitImg[fruitIndex], 1, FALSE);
        EMU_PRINT(" ghost");
	checkAndLoadTGA(&aGhostImg[ghostIndex], 7, FALSE);
      }

      // Need to delete and recreate ghosts since sprite changes every level
      for(i = 0; i < MAX_GHOSTS; i++) {
	if (ghosts[i].spr) {
	  SPRRemoveSprite(ghosts[i].spr);
	  SPRDeleteSprite(ghosts[i].spr);
	}
	ghosts[i].spr = SPRCreateSprite(&aGhostImg[ghostIndex].spr,
					i * aGhostImg[ghostIndex].w, 0,
					aGhostImg[ghostIndex].w, aGhostImg[ghostIndex].w);
      }

      EMU_PRINT("\nCreating maze...");
      // Draw the maze given the current randomly selected sprites
      CreateMaze(curLevel, &aWallImg[wallIndex], &aPillImg[pillIndex], &aFruitImg[fruitIndex],
		 MazeBuf);
      EMU_PRINT("done");
      for (i = 0, ptr = maze; i < (curLevel->w * curLevel->h); i++, ptr++) {
	if ((*ptr == TILE_START) || ((*ptr >= TILE_ENEMY_MIN) && (*ptr <= TILE_ENEMY_MAX)))
	  *ptr = TILE_EMPTY;
	if ((*ptr >= TILE_WALL_MIN) && (*ptr <= TILE_WALL_MAX))
	  *ptr = TILE_WALL;
      }

      ResetSprites();

      for(i = 0; i < NUM_BUFFERS; i++) {
	myFillRect(FrameBuffer[i], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);
	CopyMaze(FrameBuffer[i], MazeBuf, SCREEN_WIDTH * SCREEN_HEIGHT * 4 / 256);
      }

      next = chomper.aphase = 1;
      chomper.angle = 0;
      pinc = 1;

      SPRSetSpriteSource(chomper.chomperman, 
			 (chomper.aphase * SPRITESIZE) / aPlayerImg[playerIndex].scale, 0,
			 aPlayerImg[playerIndex].w, aPlayerImg[playerIndex].w);

      // Inner game play loop

      if ((gameMode == MODE_TOURNAMENT) || (gameMode == MODE_LEVEL_EDITOR)) {
	myPlayWAV((void *) GoWav, -1, &Pan, 0x40000000, 0x40000000);
      }

      for(;;) {
	// My PAUSE feature (by using start button) 1/14/2002
	if (ButtonStart(_Controller[1]))
	  if (PauseGame()) {
	    if (chase && !mute) PCMVoiceOff(voiceChase);
	    if (gameMode != MODE_LEVEL_EDITOR) gameMode = MODE_SELECT;

	    // Remove ghost sprites
	    for(i = 0; i < numGhosts; i++) SPRRemoveSprite(ghosts[i].spr);

	    // Remove smoke sprites
	    resetSmoke();

	    // Reset Chomper orientation
	    SPRSetSpriteRotation(chomper.chomperman, 0);
	    SPRSetSpriteScale(chomper.chomperman,
			      CHOMPERSCALE * aPlayerImg[playerIndex].scale,
			      CHOMPERSCALE * aPlayerImg[playerIndex].scale);

	    // Remove bonus sprites
	    if (saphase) {  SPRRemoveSprite(bonusSprite);  saphase = 0;  }
	    break;
	  }

	if ((gameMode != MODE_TOURNAMENT) && (score > extra)) {
	  myPlayWAV((void *) ExtraChomperWav, -1, &Pan, 0x40000000, 0x40000000);
	  lives++;
	  extra += 12000;
	}

	// Handle percentage
	if (gameMode == MODE_TOURNAMENT) {
	  if (percentDelay) {
	    percentDelay--;
	    if (!percentDelay && !gameover)
	      myPlayWAV((void *) ClockTickWav, -1, &Pan, 0x40000000, 0x40000000);
	  } else {
	    percent++;
	  }
	}

	if (delay) _TimeToSleep(delay);

	// Check if level done
	if (collected == pills) {
	  levelDone();
	  break;
	}

	// Are we in "power pill" mode?
	if (chase) {
	  // Fade ghost colors in & out
	  for(i = 0; i < numGhosts; i++)
	    if (ghosts[i].active && !ghosts[i].immune)
	      SPRSetSpriteType(ghosts[i].spr, 0x2+(chase%32)*0x2000000+0x2000000);
	    else
	      SPRSetSpriteType(ghosts[i].spr, 0x40000002);
	  chase--;
	  if (!chase) {
	    // Exiting "power pill" mode - return sprites to normal color
	    for(i = 0; i < numGhosts; i++) SPRSetSpriteType(ghosts[i].spr, 0x40000002);
	    if (!mute) PCMVoiceOff(voiceChase);
	  }
	}

	if (gameover) {
	  // Game over
		  
	  // This waits until the "death" sound is done
	  if (mute || PCMGetUsedVoices()) {
	    // Rotate chomperman while "dying" sound is playing
	    SPRSetSpriteRotation(chomper.chomperman, gameover * 0x800);
	    SPRSetSpriteScale(chomper.chomperman,
			      (CHOMPERSCALE -
			       ((gameover * 0x100) / 2)) * aPlayerImg[playerIndex].scale,
			      (CHOMPERSCALE -
			       ((gameover * 0x100) / 2)) * aPlayerImg[playerIndex].scale);
	    gameover++;

	  } else {
	    // Sound is done, reinit player
	    if (gameMode != MODE_TOURNAMENT) lives--;
	    gameover = 0;
	    ResetSprites();
	    if (((gameMode != MODE_TOURNAMENT) && !lives) ||
		((gameMode == MODE_TOURNAMENT) && (percent >= curLevel->distance))) {
	      drawScore(score, FrameBuffer[drawBuffer], SCOREX, SCOREY, 0x40000000, 
			(gameMode == MODE_TOURNAMENT) ? 0 : 1);
              checkAndLoadTGA(&aImage[kImgGameOver], 1, TRUE);
	      SPRBlitter(aImage[kImgGameOver].spr.img, aImage[kImgGameOver].spr.dmaflags,
                         0, 0, 256, 128,
			 FrameBuffer[drawBuffer], DMAFLAGS, SCREEN_WIDTH/2, SCREEN_HEIGHT/2,
			 0x10000 * SCREEN_WIDTH / 720, 0x10000 * SCREEN_HEIGHT / 480, 0,
			 0, 0, SCREEN_WIDTH - 1,SCREEN_HEIGHT - 1, 0x40000002, 0);
			  
	      VideoSync();
	      WaitForJoystick(JOY_A);

	      // Remove extra sprites
	      resetSmoke();
	      if (saphase) {  SPRRemoveSprite(bonusSprite);  saphase = 0;  }
	      for(i = 0; i < numGhosts; i++) SPRRemoveSprite(ghosts[i].spr);
			  
	      curLevelIndex = numLevels; // force restart
	      break;
	    }
	  }
	} else {
	  //
	  // Normal loop
	  //

	  // Move sprites
	  // SDC 2002-05-26:  Removed divider from ghost movement
	  ControlChomperman();
	  for (i = 0; i < numGhosts; i++) ControlGhost(&ghosts[i]);

	  handleSmoke();

	  collisionDetect();
	  if ((gameMode == MODE_TOURNAMENT) && (percent >= curLevel->distance)) {
	    myPlayWAV((void *) ChomperDiesWav, -1, &Pan, 0x40000000, 0x40000000);
	    gameover = 1;
            if (chase) { if (!mute) PCMVoiceOff(voiceChase);  chase = 0; }
	  }

	  if (score > 99999) score -= 99999;

	  DoScoreAnimation();

	  // SDC 2002-02-10:  Parameterized number of animation frames and increased to 4
	  // SDC 2002-02-10:  Sped up animation to 1 per frame
	  // SDC 2002-02-16:  Slowed down to every other frame
	  // SDC 2002-06-12:  Slowed down to every third frame
	  if (chomper.direction) {
	    if (!chomper.animDelay) {
	      SPRSetSpriteSource(chomper.chomperman, 
				 (chomper.aphase * SPRITESIZE) / aPlayerImg[playerIndex].scale, 0,
				 aPlayerImg[playerIndex].w, aPlayerImg[playerIndex].w);
	      chomper.aphase += pinc;
	      if ((chomper.aphase == 0) || (chomper.aphase == (CHOMPER_ANIM_FRAMES-1))) pinc = -pinc;
	      chomper.animDelay = ANIM_DELAY;
	    } else {
	      chomper.animDelay--;
	    }
	  }
	}

	VideoSync();
	      
	// Copy maze into frame buffer
	CopyMaze(FrameBuffer[drawBuffer], MazeBuf, SCREEN_WIDTH * SCREEN_HEIGHT * 4 / 256);

	// Print score   
	drawScore(score, FrameBuffer[drawBuffer], SCOREX, SCOREY, 0x40000000,
		  (gameMode == MODE_TOURNAMENT) ? 0 : 1);

	// Draw chase progress bar
	if (chase)
	  myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, CHASEBARX, CHASEBARY,
		     (150 * chase) / CHASE_TIME, TILESIZE, kGreen);
	      
	// Process display list
	SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH - 1,
			 SCREEN_HEIGHT - 1, kBlack);
	SPRDraw(0,0);

	SPRWait();
      }

    } // Rely on 'break's to exit this loop

  }

}

inline int myPlayWAV(void *wavaddr, int voice, PCMPOS *Pan, int volume, int echo)
{
  if (!mute && initAudio) return(PlayWAV(wavaddr, voice, Pan, volume, echo));
  return(-1);
}

//==================================================================================================
// checkAndLoadTGA
//
// Checks a TGA and installs it if not already done.
//
// Revision history
//   SDC 2002-02-17:  Initial revision
//==================================================================================================

void checkAndLoadTGA(sCustomImg *img, int spritesX, int emuPrint) {
  if (!img->init) {
    if (emuPrint) EMU_PRINT("\nLoading new image...");
    img->init = TRUE;
    img->transColor = getFirstTGAPixel(img->tgaAddr);
    retVal = SPRInstallTGAImage(img->tgaAddr, 1, &img->spr, img->transColor);
    setImageScale(img, spritesX);
    if (emuPrint) EMU_PRINT("done");
  }
}

